# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
# Copyright 2012 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Command-line flag library.

Emulates gflags by wrapping cfg.ConfigOpts.

The idea is to move fully to cfg eventually, and this wrapper is a
stepping stone.

"""

import socket

from oslo_config import cfg
from oslo_log import log
from oslo_middleware import cors
from oslo_utils import netutils
import six

from manila.common import constants
from manila import exception
from manila.i18n import _

CONF = cfg.CONF
log.register_options(CONF)


core_opts = [
    cfg.StrOpt('state_path',
               default='/var/lib/manila',
               help="Top-level directory for maintaining manila's state."),
]

debug_opts = [
]

CONF.register_cli_opts(core_opts)
CONF.register_cli_opts(debug_opts)

global_opts = [
    cfg.HostAddressOpt('my_ip',
                       default=netutils.get_my_ipv4(),
                       sample_default='<your_ip>',
                       help='IP address of this host.'),
    cfg.StrOpt('scheduler_topic',
               default='manila-scheduler',
               help='The topic scheduler nodes listen on.'),
    cfg.StrOpt('share_topic',
               default='manila-share',
               help='The topic share nodes listen on.'),
    cfg.StrOpt('data_topic',
               default='manila-data',
               help='The topic data nodes listen on.'),
    cfg.BoolOpt('api_rate_limit',
                default=True,
                help='Whether to rate limit the API.'),
    cfg.ListOpt('osapi_share_ext_list',
                default=[],
                help='Specify list of extensions to load when using osapi_'
                     'share_extension option with manila.api.contrib.'
                     'select_extensions.'),
    cfg.ListOpt('osapi_share_extension',
                default=['manila.api.contrib.standard_extensions'],
                help='The osapi share extensions to load.'),
    cfg.StrOpt('scheduler_manager',
               default='manila.scheduler.manager.SchedulerManager',
               help='Full class name for the scheduler manager.'),
    cfg.StrOpt('share_manager',
               default='manila.share.manager.ShareManager',
               help='Full class name for the share manager.'),
    cfg.StrOpt('data_manager',
               default='manila.data.manager.DataManager',
               help='Full class name for the data manager.'),
    cfg.HostAddressOpt('host',
                       default=socket.gethostname(),
                       sample_default='<your_hostname>',
                       help='Name of this node.  This can be an opaque '
                            'identifier. It is not necessarily a hostname, '
                            'FQDN, or IP address.'),
    # NOTE(vish): default to nova for compatibility with nova installs
    cfg.StrOpt('storage_availability_zone',
               default='nova',
               help='Availability zone of this node.'),
    cfg.StrOpt('default_share_type',
               help='Default share type to use.'),
    cfg.StrOpt('default_share_group_type',
               help='Default share group type to use.'),
    cfg.ListOpt('memcached_servers',
                help='Memcached servers or None for in process cache.',
                deprecated_reason="The config option is not used. It should "
                "not be confused with [keystone_authtoken]/memcached_servers.",
                deprecated_for_removal=True),
    cfg.StrOpt('share_usage_audit_period',
               default='month',
               deprecated_for_removal=True,
               help='Time period to generate share usages for.  '
                    'Time period must be hour, day, month or year.',
               deprecated_reason="The config option is not used."),
    cfg.StrOpt('rootwrap_config',
               help='Path to the rootwrap configuration file to use for '
                    'running commands as root.'),
    cfg.BoolOpt('monkey_patch',
                default=False,
                help='Whether to log monkey patching.'),
    cfg.ListOpt('monkey_patch_modules',
                default=[],
                help='List of modules or decorators to monkey patch.'),
    cfg.IntOpt('service_down_time',
               default=60,
               help='Maximum time since last check-in for up service.'),
    cfg.StrOpt('share_api_class',
               default='manila.share.api.API',
               help='The full class name of the share API class to use.'),
    cfg.StrOpt('auth_strategy',
               default='keystone',
               help='The strategy to use for auth. Supports noauth, keystone, '
                    'and deprecated.'),
    cfg.ListOpt('enabled_share_backends',
                help='A list of share backend names to use. These backend '
                     'names should be backed by a unique [CONFIG] group '
                     'with its options.'),
    cfg.ListOpt('enabled_share_protocols',
                default=['NFS', 'CIFS'],
                help="Specify list of protocols to be allowed for share "
                     "creation. Available values are '%s'" % six.text_type(
                         constants.SUPPORTED_SHARE_PROTOCOLS)),
]

CONF.register_opts(global_opts)


def verify_share_protocols():
    """Perform verification of 'enabled_share_protocols'."""
    msg = None
    supported_protocols = constants.SUPPORTED_SHARE_PROTOCOLS
    data = dict(supported=', '.join(supported_protocols))
    if CONF.enabled_share_protocols:
        for share_proto in CONF.enabled_share_protocols:
            if share_proto.upper() not in supported_protocols:
                data.update({'share_proto': share_proto})
                msg = ("Unsupported share protocol '%(share_proto)s' "
                       "is set as enabled. Available values are "
                       "%(supported)s. ")
                break
    else:
        msg = ("No share protocols were specified as enabled. "
               "Available values are %(supported)s. ")
    if msg:
        msg += ("Please specify one or more protocols using "
                "configuration option 'enabled_share_protocols'.")
        # NOTE(vponomaryov): use translation to unicode explicitly,
        # because of 'lazy' translations.
        msg = six.text_type(_(msg) % data)  # noqa H701
        raise exception.ManilaException(message=msg)


def set_middleware_defaults():
    """Update default configuration options for oslo.middleware."""
    cors.set_defaults(
        allow_headers=['X-Auth-Token',
                       'X-OpenStack-Request-ID',
                       'X-Openstack-Manila-Api-Version',
                       'X-OpenStack-Manila-API-Experimental',
                       'X-Identity-Status',
                       'X-Roles',
                       'X-Service-Catalog',
                       'X-User-Id',
                       'X-Tenant-Id'],
        expose_headers=['X-Auth-Token',
                        'X-OpenStack-Request-ID',
                        'X-Openstack-Manila-Api-Version',
                        'X-OpenStack-Manila-API-Experimental',
                        'X-Subject-Token',
                        'X-Service-Token'],
        allow_methods=['GET',
                       'PUT',
                       'POST',
                       'DELETE',
                       'PATCH']
    )



# Copyright (c) 2015 Clinton Knight.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
NetApp cDOT NFS protocol helper class.
"""

import uuid

from oslo_log import log
import six

from manila.common import constants
from manila import exception
from manila.share.drivers.netapp.dataontap.protocols import base
from manila.share.drivers.netapp import utils as na_utils


LOG = log.getLogger(__name__)


class NetAppCmodeNFSHelper(base.NetAppBaseHelper):
    """NetApp cDOT NFS protocol helper class."""

    @staticmethod
    def _escaped_address(address):
        if ':' in address:
            return '[%s]' % address
        else:
            return address

    @na_utils.trace
    def create_share(self, share, share_name,
                     clear_current_export_policy=True):
        """Creates NFS share."""
        if clear_current_export_policy:
            self._client.clear_nfs_export_policy_for_volume(share_name)
        self._ensure_export_policy(share, share_name)
        export_path = self._client.get_volume_junction_path(share_name)

        # Return a callback that may be used for generating export paths
        # for this share.
        return (lambda export_address, export_path=export_path:
                ':'.join([self._escaped_address(export_address),
                          export_path]))

    @na_utils.trace
    @base.access_rules_synchronized
    def delete_share(self, share, share_name):
        """Deletes NFS share."""
        LOG.debug('Deleting NFS export policy for share %s', share['id'])
        export_policy_name = self._get_export_policy_name(share)
        self._client.clear_nfs_export_policy_for_volume(share_name)
        self._client.soft_delete_nfs_export_policy(export_policy_name)

    @na_utils.trace
    @base.access_rules_synchronized
    def update_access(self, share, share_name, rules):
        """Replaces the list of access rules known to the backend storage."""

        # Ensure rules are valid
        for rule in rules:
            self._validate_access_rule(rule)

        # Sort rules by ascending network size
        new_rules = {rule['access_to']: rule['access_level'] for rule in rules}
        addresses = sorted(new_rules, reverse=True)

        # Ensure current export policy has the name we expect
        self._ensure_export_policy(share, share_name)
        export_policy_name = self._get_export_policy_name(share)

        # Make temp policy names so this non-atomic workflow remains resilient
        # across process interruptions.
        temp_new_export_policy_name = self._get_temp_export_policy_name()
        temp_old_export_policy_name = self._get_temp_export_policy_name()

        # Create new export policy
        self._client.create_nfs_export_policy(temp_new_export_policy_name)

        # Add new rules to new policy
        for address in addresses:
            self._client.add_nfs_export_rule(
                temp_new_export_policy_name, address,
                self._is_readonly(new_rules[address]))

        # Rename policy currently in force
        LOG.info('Renaming NFS export policy for share %(share)s to '
                 '%(policy)s.',
                 {'share': share_name, 'policy': temp_old_export_policy_name})
        self._client.rename_nfs_export_policy(export_policy_name,
                                              temp_old_export_policy_name)

        # Switch share to the new policy
        LOG.info('Setting NFS export policy for share %(share)s to '
                 '%(policy)s.',
                 {'share': share_name, 'policy': temp_new_export_policy_name})
        self._client.set_nfs_export_policy_for_volume(
            share_name, temp_new_export_policy_name)

        # Delete old policy
        self._client.soft_delete_nfs_export_policy(temp_old_export_policy_name)

        # Rename new policy to its final name
        LOG.info('Renaming NFS export policy for share %(share)s to '
                 '%(policy)s.',
                 {'share': share_name, 'policy': export_policy_name})
        self._client.rename_nfs_export_policy(temp_new_export_policy_name,
                                              export_policy_name)

    @na_utils.trace
    def _validate_access_rule(self, rule):
        """Checks whether access rule type and level are valid."""

        if rule['access_type'] != 'ip':
            msg = ("Clustered Data ONTAP supports only 'ip' type for share "
                   "access rules with NFS protocol.")
            raise exception.InvalidShareAccess(reason=msg)

        if rule['access_level'] not in constants.ACCESS_LEVELS:
            raise exception.InvalidShareAccessLevel(level=rule['access_level'])

    @na_utils.trace
    def get_target(self, share):
        """Returns ID of target OnTap device based on export location."""
        return self._get_export_location(share)[0]

    @na_utils.trace
    def get_share_name_for_share(self, share):
        """Returns the flexvol name that hosts a share."""
        _, volume_junction_path = self._get_export_location(share)
        volume = self._client.get_volume_at_junction_path(volume_junction_path)
        return volume.get('name') if volume else None

    @staticmethod
    def _get_export_location(share):
        """Returns IP address and export location of an NFS share."""
        export_location = share['export_location'] or ':'
        result = export_location.rsplit(':', 1)
        if len(result) != 2:
            return ['', '']
        return result

    @staticmethod
    def _get_temp_export_policy_name():
        """Builds export policy name for an NFS share."""
        return 'temp_' + six.text_type(uuid.uuid1()).replace('-', '_')

    @staticmethod
    def _get_export_policy_name(share):
        """Builds export policy name for an NFS share."""
        return 'policy_' + share['id'].replace('-', '_')

    @na_utils.trace
    def _ensure_export_policy(self, share, share_name):
        """Ensures a flexvol/share has an export policy.

        This method ensures a flexvol has an export policy with a name
        containing the share ID.  For legacy reasons, this may not
        always be the case.
        """
        expected_export_policy = self._get_export_policy_name(share)
        actual_export_policy = self._client.get_nfs_export_policy_for_volume(
            share_name)

        if actual_export_policy == expected_export_policy:
            return
        elif actual_export_policy == 'default':
            self._client.create_nfs_export_policy(expected_export_policy)
            self._client.set_nfs_export_policy_for_volume(
                share_name, expected_export_policy)
        else:
            self._client.rename_nfs_export_policy(actual_export_policy,
                                                  expected_export_policy)



# Copyright 2017 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock
from oslo_config import cfg

from manila import exception
from manila import quota
from manila import test

CONF = cfg.CONF


@ddt.ddt
class DbQuotaDriverTestCase(test.TestCase):

    def setUp(self):
        super(DbQuotaDriverTestCase, self).setUp()
        self.project_id = 'fake_project_id'
        self.user_id = 'fake_user_id'
        self.share_type_id = 'fake_share_type_id'
        self.ctxt = type(
            'FakeContext', (object, ),
            {'project_id': self.project_id, 'user_id': self.user_id,
             'quota_class': 'fake_quota_class', 'elevated': mock.Mock()})
        self.driver = quota.DbQuotaDriver()
        self.reservations = ['foo', 'bar']
        self.resources = {k: quota.BaseResource(k) for k in ('foo', 'bar')}

    def test_get_by_class(self):
        self.mock_object(quota.db, 'quota_class_get')

        result = self.driver.get_by_class(
            self.ctxt, 'fake_quota_class', 'fake_res')

        self.assertEqual(quota.db.quota_class_get.return_value, result)
        quota.db.quota_class_get.assert_called_once_with(
            self.ctxt, 'fake_quota_class', 'fake_res')

    def test_get_defaults(self):
        self.mock_object(
            quota.db, 'quota_class_get_default',
            mock.Mock(return_value={'foo': 13}))

        result = self.driver.get_defaults(self.ctxt, self.resources)

        self.assertEqual(
            {'foo': 13, 'bar': self.resources['bar'].default}, result)
        quota.db.quota_class_get_default.assert_called_once_with(self.ctxt)

    @ddt.data(True, False)
    def test_get_class_quotas(self, defaults):
        self.mock_object(
            quota.db, 'quota_class_get_all_by_name',
            mock.Mock(return_value={'foo': 13}))

        result = self.driver.get_class_quotas(
            self.ctxt, self.resources, 'fake_quota_class', defaults)

        expected = {'foo': 13, 'bar': -1} if defaults else {'foo': 13}
        self.assertEqual(expected, result)
        quota.db.quota_class_get_all_by_name.assert_called_once_with(
            self.ctxt, 'fake_quota_class')

    @ddt.data(
        ('fake_project_id', {'foo': 20}, None, True, None, True),
        ('fake_different_project_id', {'bar': 40}, 'fake_quota_class', True,
         {'foo': {'in_use': 3, 'reserved': 2}}, False),
        ('fake_project_id', {'bar': 30}, 'fake_quota_class', True, None, False)
    )
    @ddt.unpack
    def test__process_quotas(self, project_id, quotas, quota_class, defaults,
                             usages, remains):
        self.mock_object(quota.db, 'quota_get_all', mock.Mock(return_value=[]))
        self.mock_object(quota.db, 'quota_class_get_all_by_name')
        self.mock_object(
            self.driver, 'get_defaults',
            mock.Mock(return_value={'foo': 11, 'bar': 12}))
        self.mock_object(
            quota.db, 'quota_get_all',
            mock.Mock(return_value=[]))

        result = self.driver._process_quotas(
            self.ctxt, self.resources, project_id, quotas, quota_class,
            defaults, usages, remains)

        expected = {key: {'limit': mock.ANY} for key in ('foo', 'bar')}
        if usages:
            for res in self.resources.values():
                usage = usages.get(res.name, {})
                expected[res.name].update(
                    in_use=usage.get('in_use', 0),
                    reserved=usage.get('reserved', 0))
        if remains:
            quota.db.quota_get_all.assert_called_once_with(
                self.ctxt, project_id)
            for res in self.resources.values():
                expected[res.name]['remains'] = mock.ANY
        else:
            self.assertEqual(0, quota.db.quota_get_all.call_count)

        self.assertEqual(expected, result)
        if quota_class or project_id == self.ctxt.project_id:
            quota.db.quota_class_get_all_by_name.assert_called_once_with(
                self.ctxt, quota_class or self.ctxt.quota_class)
        else:
            self.assertEqual(
                0, quota.db.quota_class_get_all_by_name.call_count)

    @ddt.data(
        ('fake_quota_class', True, None, 'fake_remains'),
        (None, False, 'fake_usages', False),
    )
    @ddt.unpack
    def test_get_project_quotas(self, quota_class, defaults, usages, remains):
        self.mock_object(quota.db, 'quota_get_all_by_project')
        self.mock_object(quota.db, 'quota_usage_get_all_by_project')
        self.mock_object(self.driver, '_process_quotas')

        result = self.driver.get_project_quotas(
            self.ctxt, self.resources, self.project_id,
            quota_class, defaults, usages, remains)

        self.assertEqual(
            result, self.driver._process_quotas.return_value)
        project_usages = None
        if usages:
            project_usages = (
                quota.db.quota_usage_get_all_by_project.return_value)
        self.driver._process_quotas.assert_called_once_with(
            self.ctxt, self.resources, self.project_id,
            quota.db.quota_get_all_by_project.return_value,
            quota_class, defaults=defaults, usages=project_usages,
            remains=remains)
        quota.db.quota_get_all_by_project.assert_called_once_with(
            self.ctxt, self.project_id)
        if usages:
            quota.db.quota_usage_get_all_by_project.assert_called_once_with(
                self.ctxt, self.project_id)
        else:
            self.assertEqual(
                0, quota.db.quota_usage_get_all_by_project.call_count)

    @ddt.data(
        (None, True, True),
        ('fake_quota_class', False, True),
        ('fake_quota_class', True, False),
    )
    @ddt.unpack
    def test_get_user_quotas(self, quota_class, defaults, usages):
        project_quotas = {'fake_resource': 5}
        self.mock_object(
            quota.db, 'quota_get_all_by_project',
            mock.Mock(return_value=project_quotas))
        self.mock_object(
            quota.db, 'quota_get_all_by_project_and_user',
            mock.Mock(return_value={'fake_user_defined_resource': 14}))
        mock_user_usages = self.mock_object(
            quota.db, 'quota_usage_get_all_by_project_and_user')
        self.mock_object(self.driver, '_process_quotas')

        result = self.driver.get_user_quotas(
            self.ctxt, self.resources, self.project_id, self.user_id,
            quota_class, defaults, usages)

        self.assertEqual(
            self.driver._process_quotas.return_value, result)
        quota.db.quota_get_all_by_project.assert_called_once_with(
            self.ctxt, self.project_id)
        quota.db.quota_get_all_by_project_and_user.assert_called_once_with(
            self.ctxt, self.project_id, self.user_id)
        if usages:
            user_usages = mock_user_usages.return_value
            mock_user_usages.assert_called_once_with(
                self.ctxt, self.project_id, self.user_id)
        else:
            user_usages = None
            self.assertEqual(0, mock_user_usages.call_count)
        expected_user_quotas = {'fake_user_defined_resource': 14}
        expected_user_quotas.update(project_quotas)
        self.driver._process_quotas.assert_called_once_with(
            self.ctxt, self.resources, self.project_id, expected_user_quotas,
            quota_class, defaults=defaults, usages=user_usages)

    @ddt.data(
        (None, True, True),
        ('fake_quota_class', False, True),
        ('fake_quota_class', True, False),
    )
    @ddt.unpack
    def test_get_share_type_quotas(self, quota_class, defaults, usages):
        project_quotas = {'fake_resource': 5}
        self.mock_object(
            quota.db, 'quota_get_all_by_project',
            mock.Mock(return_value=project_quotas))
        mock_st_quotas = self.mock_object(
            quota.db, 'quota_get_all_by_project_and_share_type',
            mock.Mock(return_value={'fake_st_defined_resource': 14}))
        mock_st_usages = self.mock_object(
            quota.db, 'quota_usage_get_all_by_project_and_share_type')
        self.mock_object(self.driver, '_process_quotas')

        result = self.driver.get_share_type_quotas(
            self.ctxt, self.resources, self.project_id, self.share_type_id,
            quota_class, defaults, usages)

        self.assertEqual(
            self.driver._process_quotas.return_value, result)
        quota.db.quota_get_all_by_project.assert_called_once_with(
            self.ctxt, self.project_id)
        mock_st_quotas.assert_called_once_with(
            self.ctxt, self.project_id, self.share_type_id)
        if usages:
            st_usages = mock_st_usages.return_value
            mock_st_usages.assert_called_once_with(
                self.ctxt, self.project_id, self.share_type_id)
        else:
            st_usages = None
            self.assertEqual(0, mock_st_usages.call_count)
        expected_st_quotas = {'fake_st_defined_resource': 14}
        expected_st_quotas.update(project_quotas)
        self.driver._process_quotas.assert_called_once_with(
            self.ctxt, self.resources, self.project_id, expected_st_quotas,
            quota_class, defaults=defaults, usages=st_usages)

    @ddt.data((None, None), (None, 'foo_st_id'), ('foo_user_id', None))
    @ddt.unpack
    def test_get_settable_quotas(self, user_id, st_id):
        project_quotas = {'fake': {
            'limit': 13, 'in_use': 7, 'reserved': 5, 'remains': 1,
        }}
        user_or_st_quotas = {'fake': {
            'limit': 11, 'in_use': 5, 'reserved': 2,
        }}
        self.mock_object(
            self.driver, 'get_project_quotas',
            mock.Mock(return_value=project_quotas))
        self.mock_object(
            self.driver, 'get_user_quotas',
            mock.Mock(return_value=user_or_st_quotas))
        self.mock_object(
            self.driver, 'get_share_type_quotas',
            mock.Mock(return_value=user_or_st_quotas))

        result = self.driver.get_settable_quotas(
            self.ctxt, self.resources, self.project_id, user_id, st_id)

        if user_id:
            self.driver.get_user_quotas.assert_called_once_with(
                self.ctxt, self.resources, self.project_id, user_id)
        else:
            self.assertEqual(0, self.driver.get_user_quotas.call_count)
        if st_id:
            self.driver.get_share_type_quotas.assert_called_once_with(
                self.ctxt, self.resources, self.project_id, st_id)
        else:
            self.assertEqual(0, self.driver.get_share_type_quotas.call_count)
        if user_id or st_id:
            expected_settable_quotas = {'fake': {'maximum': 13, 'minimum': 7}}
        else:
            expected_settable_quotas = {'fake': {'maximum': -1, 'minimum': 12}}
        self.driver.get_project_quotas.assert_called_once_with(
            self.ctxt, self.resources, self.project_id, remains=True)
        self.assertEqual(expected_settable_quotas, result)

    @ddt.data((None, None), (None, 'fake_st_id'), ('fake_user_id', None))
    @ddt.unpack
    def test__get_quotas(self, user_id, st_id):
        quotas = {'foo': {'limit': 5}, 'bar': {'limit': 13}}
        self.mock_object(
            self.driver, 'get_project_quotas', mock.Mock(return_value=quotas))
        self.mock_object(
            self.driver, 'get_user_quotas', mock.Mock(return_value=quotas))
        self.mock_object(
            self.driver, 'get_share_type_quotas',
            mock.Mock(return_value=quotas))

        result = self.driver._get_quotas(
            self.ctxt, self.resources, ('foo', 'bar'), False,
            self.project_id, user_id, st_id)

        expected = {k: v['limit'] for k, v in quotas.items()}
        self.assertEqual(expected, result)
        sub_resources = {k: v for k, v in self.resources.items()}
        if user_id:
            self.driver.get_user_quotas.assert_called_once_with(
                self.ctxt, sub_resources, self.project_id, user_id,
                self.ctxt.quota_class, usages=False)
            self.assertEqual(0, self.driver.get_project_quotas.call_count)
            self.assertEqual(0, self.driver.get_share_type_quotas.call_count)
        elif st_id:
            self.driver.get_share_type_quotas.assert_called_once_with(
                self.ctxt, sub_resources, self.project_id, st_id,
                self.ctxt.quota_class, usages=False)
            self.assertEqual(0, self.driver.get_project_quotas.call_count)
            self.assertEqual(0, self.driver.get_user_quotas.call_count)
        else:
            self.driver.get_project_quotas.assert_called_once_with(
                self.ctxt, sub_resources, self.project_id,
                self.ctxt.quota_class, usages=False)
            self.assertEqual(0, self.driver.get_user_quotas.call_count)
            self.assertEqual(0, self.driver.get_share_type_quotas.call_count)

    def test__get_quotas_unknown(self):
        quotas = {'foo': {'limit': 5}, 'bar': {'limit': 13}}
        self.mock_object(
            self.driver, 'get_project_quotas', mock.Mock(return_value=quotas))
        self.mock_object(
            self.driver, 'get_user_quotas', mock.Mock(return_value=quotas))
        self.mock_object(
            self.driver, 'get_share_type_quotas',
            mock.Mock(return_value=quotas))

        self.assertRaises(
            exception.QuotaResourceUnknown,
            self.driver._get_quotas,
            self.ctxt, self.resources, ['foo', 'bar'], True,
            self.project_id, self.user_id, self.share_type_id)

        self.assertEqual(0, self.driver.get_project_quotas.call_count)
        self.assertEqual(0, self.driver.get_user_quotas.call_count)
        self.assertEqual(0, self.driver.get_share_type_quotas.call_count)

    @ddt.data(
        {}, {'project_id': 'fake_project'}, {'user_id': 'fake_user'},
        {'share_type_id': 'fake_share_type_id'},
    )
    def test_reserve(self, kwargs):
        self.mock_object(quota.db, 'quota_reserve')
        deltas = {'delta1': 1, 'delta2': 2}
        quotas, user_quotas, st_quotas = 'fake1', 'fake2', 'fake3'
        self.mock_object(
            self.driver, '_get_quotas', mock.Mock(
                side_effect=[quotas, user_quotas, st_quotas]))

        result = self.driver.reserve(
            self.ctxt, self.resources, deltas, None, **kwargs)

        expected_kwargs = {
            'project_id': self.ctxt.project_id,
            'user_id': self.ctxt.user_id,
            'share_type_id': None,
        }
        expected_kwargs.update(kwargs)
        st_quotas = st_quotas if kwargs.get('share_type_id') else {}
        self.assertEqual(quota.db.quota_reserve.return_value, result)
        quota.db.quota_reserve.assert_called_once_with(
            self.ctxt, self.resources, quotas, user_quotas, st_quotas,
            deltas, mock.ANY, CONF.until_refresh, CONF.max_age,
            **expected_kwargs)
        self.assertEqual(
            3 if kwargs.get('share_type_id') else 2,
            self.driver._get_quotas.call_count)

    def test_reserve_wrong_expire(self):
        self.assertRaises(
            exception.InvalidReservationExpiration,
            self.driver.reserve,
            self.ctxt, self.resources, 'fake_deltas', 'fake_expire')

    def test_commit(self):
        self.mock_object(quota.db, 'reservation_commit')

        result = self.driver.commit(
            self.ctxt, self.reservations, self.project_id, self.user_id,
            self.share_type_id)

        self.assertIsNone(result)
        quota.db.reservation_commit.assert_called_once_with(
            self.ctxt, self.reservations, project_id=self.project_id,
            user_id=self.user_id, share_type_id=self.share_type_id)

    @ddt.data(
        (None, None),
        ('fake_project_id', 'fake_user_id'),
    )
    @ddt.unpack
    def test_rollback(self, project_id, user_id):
        self.mock_object(quota.db, 'reservation_rollback')

        result = self.driver.rollback(
            self.ctxt, self.reservations, project_id, user_id,
            self.share_type_id)

        expected_project_id = project_id or self.ctxt.project_id
        expected_user_id = user_id or self.ctxt.user_id
        self.assertIsNone(result)
        quota.db.reservation_rollback.assert_called_once_with(
            self.ctxt, self.reservations, project_id=expected_project_id,
            user_id=expected_user_id, share_type_id=self.share_type_id)

    def test_usage_reset(self):
        self.mock_object(
            quota.db, 'quota_usage_update',
            mock.Mock(side_effect=[
                'foo',
                exception.QuotaUsageNotFound(project_id=self.project_id)]))

        result = self.driver.usage_reset(self.ctxt, ['foo', 'bar'])

        self.assertIsNone(result)
        quota.db.quota_usage_update.assert_has_calls([
            mock.call(
                self.ctxt.elevated.return_value, self.ctxt.project_id,
                self.ctxt.user_id, res, in_use=-1)
            for res in ('foo', 'bar')
        ])

    def test_destroy_all_by_project(self):
        self.mock_object(quota.db, 'quota_destroy_all_by_project')

        result = self.driver.destroy_all_by_project(self.ctxt, self.project_id)

        self.assertIsNone(result)
        quota.db.quota_destroy_all_by_project.assert_called_once_with(
            self.ctxt, self.project_id)

    def test_destroy_all_by_project_and_user(self):
        self.mock_object(quota.db, 'quota_destroy_all_by_project_and_user')

        result = self.driver.destroy_all_by_project_and_user(
            self.ctxt, self.project_id, self.user_id)

        self.assertIsNone(result)
        quota.db.quota_destroy_all_by_project_and_user.assert_called_once_with(
            self.ctxt, self.project_id, self.user_id)

    def test_destroy_all_by_project_and_share_type(self):
        mock_destroy_all = self.mock_object(
            quota.db, 'quota_destroy_all_by_share_type')

        result = self.driver.destroy_all_by_project_and_share_type(
            self.ctxt, self.project_id, self.share_type_id)

        self.assertIsNone(result)
        mock_destroy_all.assert_called_once_with(
            self.ctxt, self.share_type_id, project_id=self.project_id)

    def test_expire(self):
        self.mock_object(quota.db, 'reservation_expire')

        result = self.driver.expire(self.ctxt)

        self.assertIsNone(result)
        quota.db.reservation_expire.assert_called_once_with(self.ctxt)


@ddt.ddt
class QuotaEngineTestCase(test.TestCase):

    def setUp(self):
        super(QuotaEngineTestCase, self).setUp()
        self.ctxt = 'fake_context'
        self.mock_class('manila.quota.DbQuotaDriver')
        self.engine = quota.QuotaEngine()
        self.driver = self.engine._driver
        self.resources = [quota.BaseResource('foo'), quota.BaseResource('bar')]
        self.project_id = 'fake_project_id'
        self.user_id = 'fake_user_id'
        self.share_type_id = 'fake_share_type_id'
        self.quota_class = 'fake_quota_class'

    def test_register_resource(self):
        self.assertNotIn(self.resources[0].name, self.engine)
        self.engine.register_resource(self.resources[0])
        self.assertIn(self.resources[0].name, self.engine)

    def test_register_resources(self):
        for res in self.resources:
            self.assertNotIn(res.name, self.engine)
        self.engine.register_resources(self.resources)
        for res in self.resources:
            self.assertIn(res.name, self.engine)

    def test_get_by_class(self):
        result = self.engine.get_by_class(
            self.ctxt, self.quota_class, 'fake_res')

        self.assertEqual(result, self.driver.get_by_class.return_value)
        self.driver.get_by_class.assert_called_once_with(
            self.ctxt, self.quota_class, 'fake_res')

    def test_get_defaults(self):
        result = self.engine.get_defaults(self.ctxt)

        self.assertEqual(result, self.driver.get_defaults.return_value)
        self.driver.get_defaults.assert_called_once_with(
            self.ctxt, self.engine._resources)

    @ddt.data(None, True, False)
    def test_get_class_quotas(self, defaults):
        kwargs = {}
        if defaults is not None:
            kwargs['defaults'] = defaults

        result = self.engine.get_class_quotas(
            self.ctxt, self.quota_class, **kwargs)

        self.assertEqual(result, self.driver.get_class_quotas.return_value)
        kwargs['defaults'] = defaults if defaults is not None else True
        self.driver.get_class_quotas.assert_called_once_with(
            self.ctxt, self.engine._resources, self.quota_class, **kwargs)

    @ddt.data(
        {},
        {'quota_class': 'foo'},
        {'defaults': False},
        {'usages': False},
    )
    def test_get_user_quotas(self, kwargs):
        expected_kwargs = {
            'quota_class': None,
            'defaults': True,
            'usages': True,
        }
        expected_kwargs.update(kwargs)

        result = self.engine.get_user_quotas(
            self.ctxt, self.project_id, self.user_id, **kwargs)

        self.assertEqual(result, self.driver.get_user_quotas.return_value)
        self.driver.get_user_quotas.assert_called_once_with(
            self.ctxt, self.engine._resources,
            self.project_id, self.user_id, **expected_kwargs)

    @ddt.data(
        {},
        {'quota_class': 'foo'},
        {'defaults': False},
        {'usages': False},
    )
    def test_get_share_type_quotas(self, kwargs):
        expected_kwargs = {
            'quota_class': None,
            'defaults': True,
            'usages': True,
        }
        expected_kwargs.update(kwargs)

        result = self.engine.get_share_type_quotas(
            self.ctxt, self.project_id, self.share_type_id, **kwargs)

        self.assertEqual(
            result, self.driver.get_share_type_quotas.return_value)
        self.driver.get_share_type_quotas.assert_called_once_with(
            self.ctxt, self.engine._resources,
            self.project_id, self.share_type_id, **expected_kwargs)

    @ddt.data(
        {},
        {'quota_class': 'foo'},
        {'defaults': False},
        {'usages': False},
        {'remains': True},
    )
    def test_get_project_quotas(self, kwargs):
        expected_kwargs = {
            'quota_class': None,
            'defaults': True,
            'usages': True,
            'remains': False,
        }
        expected_kwargs.update(kwargs)

        result = self.engine.get_project_quotas(
            self.ctxt, self.project_id, **kwargs)

        self.assertEqual(result, self.driver.get_project_quotas.return_value)
        self.driver.get_project_quotas.assert_called_once_with(
            self.ctxt, self.engine._resources,
            self.project_id, **expected_kwargs)

    @ddt.data(
        {},
        {'user_id': 'fake_user_id'},
        {'share_type_id': 'fake_share_type_id'},
    )
    def test_get_settable_quotas(self, kwargs):
        expected_kwargs = {'user_id': None, 'share_type_id': None}
        expected_kwargs.update(kwargs)

        result = self.engine.get_settable_quotas(
            self.ctxt, self.project_id, **kwargs)

        self.assertEqual(result, self.driver.get_settable_quotas.return_value)
        self.driver.get_settable_quotas.assert_called_once_with(
            self.ctxt, self.engine._resources,
            self.project_id, **expected_kwargs)

    def test_count(self):
        mock_count = mock.Mock()
        resource = quota.CountableResource('FakeCountableResource', mock_count)
        self.engine.register_resource(resource)

        result = self.engine.count(self.ctxt, resource.name)

        self.assertEqual(mock_count.return_value, result)

    def test_count_unknown_resource(self):
        self.assertRaises(
            exception.QuotaResourceUnknown,
            self.engine.count,
            self.ctxt, 'nonexistent_resource', 'foo_arg', foo='kwarg')

    def test_reserve(self):
        result = self.engine.reserve(
            self.ctxt, 'fake_expire', self.project_id, self.user_id,
            self.share_type_id, delta1=1, delta2=2)

        self.assertEqual(self.driver.reserve.return_value, result)
        self.driver.reserve.assert_called_once_with(
            self.ctxt, self.engine._resources, {'delta1': 1, 'delta2': 2},
            expire='fake_expire', project_id=self.project_id,
            user_id=self.user_id, share_type_id=self.share_type_id)

    @ddt.data(Exception('FakeException'), [None])
    def test_commit(self, side_effect):
        fake_reservations = ['foo', 'bar']
        self.driver.commit.side_effect = side_effect
        self.mock_object(quota.LOG, 'exception')

        result = self.engine.commit(
            self.ctxt, fake_reservations, 'fake_project_id',
            'fake_user_id', 'fake_share_type_id')

        self.assertIsNone(result)
        self.driver.commit.assert_called_once_with(
            self.ctxt, fake_reservations, project_id='fake_project_id',
            user_id='fake_user_id', share_type_id='fake_share_type_id')
        if side_effect == [None]:
            self.assertEqual(0, quota.LOG.exception.call_count)
        else:
            quota.LOG.exception.assert_called_once_with(
                mock.ANY, fake_reservations)

    @ddt.data(Exception('FakeException'), [None])
    def test_rollback(self, side_effect):
        fake_reservations = ['foo', 'bar']
        self.driver.rollback.side_effect = side_effect
        self.mock_object(quota.LOG, 'exception')

        result = self.engine.rollback(
            self.ctxt, fake_reservations, 'fake_project_id',
            'fake_user_id', 'fake_share_type_id')

        self.assertIsNone(result)
        self.driver.rollback.assert_called_once_with(
            self.ctxt, fake_reservations, project_id='fake_project_id',
            user_id='fake_user_id', share_type_id='fake_share_type_id')
        if side_effect == [None]:
            self.assertEqual(0, quota.LOG.exception.call_count)
        else:
            quota.LOG.exception.assert_called_once_with(
                mock.ANY, fake_reservations)

    def test_usage_reset(self):
        result = self.engine.usage_reset(self.ctxt, 'fake_resources')

        self.assertIsNone(result)
        self.driver.usage_reset.assert_called_once_with(
            self.ctxt, 'fake_resources')

    def test_destroy_all_by_project_and_user(self):
        result = self.engine.destroy_all_by_project_and_user(
            self.ctxt, 'fake_project_id', 'fake_user_id')

        self.assertIsNone(result)
        self.driver.destroy_all_by_project_and_user.assert_called_once_with(
            self.ctxt, 'fake_project_id', 'fake_user_id')

    def test_destroy_all_by_project_and_share_type(self):
        result = self.engine.destroy_all_by_project_and_share_type(
            self.ctxt, 'fake_project_id', 'fake_st_id')

        self.assertIsNone(result)
        mock_destroy_all_by_project_and_share_type = (
            self.driver.destroy_all_by_project_and_share_type)
        mock_destroy_all_by_project_and_share_type.assert_called_once_with(
            self.ctxt, 'fake_project_id', 'fake_st_id')

    def test_destroy_all_by_project(self):
        result = self.engine.destroy_all_by_project(
            self.ctxt, 'fake_project_id')

        self.assertIsNone(result)
        self.driver.destroy_all_by_project.assert_called_once_with(
            self.ctxt, 'fake_project_id')

    def test_expire(self):
        result = self.engine.expire(self.ctxt)

        self.assertIsNone(result)
        self.driver.expire.assert_called_once_with(self.ctxt)

    def test_resources(self):
        self.engine.register_resources(self.resources)
        self.assertEqual(['bar', 'foo'], self.engine.resources)

    def test_current_common_resources(self):
        self.assertEqual(
            ['gigabytes', 'share_group_snapshots', 'share_groups',
             'share_networks', 'shares', 'snapshot_gigabytes', 'snapshots'],
            quota.QUOTAS.resources)



# Copyright (c) 2015 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import errno
import os

import ddt
import mock
from oslo_config import cfg
from oslo_utils import importutils

from manila import exception
from manila.share import configuration as config
from manila.share import driver
from manila.share.drivers.glusterfs import layout
from manila import test
from manila.tests import fake_share
from manila.tests import fake_utils


CONF = cfg.CONF


fake_local_share_path = '/mnt/nfs/testvol/fakename'

fake_path_to_private_key = '/fakepath/to/privatekey'
fake_remote_server_password = 'fakepassword'


def fake_access(kwargs):
    fake_access_rule = fake_share.fake_access(**kwargs)
    fake_access_rule.to_dict = lambda: fake_access_rule.values
    return fake_access_rule


class GlusterfsFakeShareDriver(layout.GlusterfsShareDriverBase):

    supported_layouts = ('layout_fake.FakeLayout',
                         'layout_something.SomeLayout')
    supported_protocols = ('NFS,')
    _supported_access_types = ('ip',)
    _supported_access_levels = ('rw',)


@ddt.ddt
class GlusterfsShareDriverBaseTestCase(test.TestCase):
    """Tests GlusterfsShareDriverBase."""

    def setUp(self):
        super(GlusterfsShareDriverBaseTestCase, self).setUp()
        CONF.set_default('driver_handles_share_servers', False)
        fake_conf, __ = self._setup()
        self._driver = GlusterfsFakeShareDriver(False, configuration=fake_conf)
        self.fake_share = mock.Mock(name='fake_share')
        self.fake_context = mock.Mock(name='fake_context')
        self.fake_access = mock.Mock(name='fake_access')

    def _setup(self):
        fake_conf = config.Configuration(None)
        fake_layout = mock.Mock()
        self.mock_object(importutils, "import_object",
                         mock.Mock(return_value=fake_layout))
        return fake_conf, fake_layout

    def test_init(self):
        self.assertRaises(IndexError, layout.GlusterfsShareDriverBase, False,
                          configuration=config.Configuration(None))

    @ddt.data({'has_snap': None, 'layout_name': None},
              {'has_snap': False, 'layout_name': 'layout_fake.FakeLayout'},
              {'has_snap': True, 'layout_name': 'layout_something.SomeLayout'})
    @ddt.unpack
    def test_init_subclass(self, has_snap, layout_name):
        conf, _layout = self._setup()
        if layout_name is not None:
            conf.glusterfs_share_layout = layout_name
        if has_snap is None:
            del(_layout._snapshots_are_supported)
        else:
            _layout._snapshots_are_supported = has_snap

        _driver = GlusterfsFakeShareDriver(False, configuration=conf)

        snap_result = {None: False}.get(has_snap, has_snap)
        layout_result = {None: 'layout_fake.FakeLayout'}.get(layout_name,
                                                             layout_name)

        importutils.import_object.assert_called_once_with(
            'manila.share.drivers.glusterfs.%s' % layout_result,
            _driver, configuration=conf)
        self.assertEqual(_layout, _driver.layout)
        self.assertEqual(snap_result, _driver.snapshots_are_supported)

    def test_init_nosupp_layout(self):
        conf = config.Configuration(None)
        conf.glusterfs_share_layout = 'nonsense_layout'

        self.assertRaises(exception.GlusterfsException,
                          GlusterfsFakeShareDriver, False, configuration=conf)

    def test_setup_via_manager(self):
        self.assertIsNone(self._driver._setup_via_manager(mock.Mock()))

    def test_supported_access_types(self):
        self.assertEqual(('ip',), self._driver.supported_access_types)

    def test_supported_access_levels(self):
        self.assertEqual(('rw',), self._driver.supported_access_levels)

    def test_access_rule_validator(self):
        rule = mock.Mock()
        abort = mock.Mock()
        valid = mock.Mock()
        self.mock_object(layout.ganesha_utils, 'validate_access_rule',
                         mock.Mock(return_value=valid))

        ret = self._driver._access_rule_validator(abort)(rule)

        self.assertEqual(valid, ret)
        layout.ganesha_utils.validate_access_rule.assert_called_once_with(
            ('ip',), ('rw',), rule, abort)

    @ddt.data({'inset': ([], ['ADD'], []), 'outset': (['ADD'], []),
               'recovery': False},
              {'inset': ([], [], ['DELETE']), 'outset': ([], ['DELETE']),
               'recovery': False},
              {'inset': (['EXISTING'], ['ADD'], ['DELETE']),
               'outset': (['ADD'], ['DELETE']), 'recovery': False},
              {'inset': (['EXISTING'], [], []), 'outset': (['EXISTING'], []),
               'recovery': True})
    @ddt.unpack
    def test_update_access(self, inset, outset, recovery):
        conf, _layout = self._setup()
        gluster_mgr = mock.Mock(name='gluster_mgr')
        self.mock_object(_layout, '_share_manager',
                         mock.Mock(return_value=gluster_mgr))
        _driver = GlusterfsFakeShareDriver(False, configuration=conf)
        self.mock_object(_driver, '_update_access_via_manager', mock.Mock())
        rulemap = {t: fake_access({'access_type': "ip",
                                   'access_level': "rw",
                                   'access_to': t}) for t in (
            'EXISTING', 'ADD', 'DELETE')}
        in_rules, out_rules = (
            [
                [
                    rulemap[t] for t in r
                ] for r in rs
            ] for rs in (inset, outset))

        _driver.update_access(self.fake_context, self.fake_share, *in_rules)

        _layout._share_manager.assert_called_once_with(self.fake_share)
        _driver._update_access_via_manager.assert_called_once_with(
            gluster_mgr, self.fake_context, self.fake_share,
            *out_rules, recovery=recovery)

    def test_update_access_via_manager(self):
        self.assertRaises(NotImplementedError,
                          self._driver._update_access_via_manager,
                          mock.Mock(), self.fake_context, self.fake_share,
                          [self.fake_access], [self.fake_access])

    @ddt.data('NFS', 'PROTATO')
    def test_check_proto_baseclass(self, proto):
        self.assertRaises(exception.ShareBackendException,
                          layout.GlusterfsShareDriverBase._check_proto,
                          {'share_proto': proto})

    def test_check_proto(self):
        GlusterfsFakeShareDriver._check_proto({'share_proto': 'NFS'})

    def test_check_proto_notsupported(self):
        self.assertRaises(exception.ShareBackendException,
                          GlusterfsFakeShareDriver._check_proto,
                          {'share_proto': 'PROTATO'})

    @ddt.data('', '_from_snapshot')
    def test_create_share(self, variant):
        conf, _layout = self._setup()
        _driver = GlusterfsFakeShareDriver(False, configuration=conf)
        self.mock_object(_driver, '_check_proto', mock.Mock())

        getattr(_driver, 'create_share%s' % variant)(self.fake_context,
                                                     self.fake_share)

        _driver._check_proto.assert_called_once_with(self.fake_share)
        getattr(_layout,
                'create_share%s' % variant).assert_called_once_with(
            self.fake_context, self.fake_share)

    @ddt.data(True, False)
    def test_update_share_stats(self, internal_exception):
        data = mock.Mock()
        conf, _layout = self._setup()

        def raise_exception(*args, **kwargs):
            raise NotImplementedError
        layoutstats = mock.Mock()

        mock_kw = ({'side_effect': raise_exception} if internal_exception
                   else {'return_value': layoutstats})

        self.mock_object(_layout, '_update_share_stats', mock.Mock(**mock_kw))
        self.mock_object(driver.ShareDriver, '_update_share_stats',
                         mock.Mock())
        _driver = GlusterfsFakeShareDriver(False, configuration=conf)
        _driver._update_share_stats(data)

        if internal_exception:
            self.assertFalse(data.update.called)
        else:
            data.update.assert_called_once_with(layoutstats)
        driver.ShareDriver._update_share_stats.assert_called_once_with(
            data)

    @ddt.data('do_setup', 'create_snapshot', 'delete_share', 'delete_snapshot',
              'ensure_share', 'manage_existing', 'unmanage', 'extend_share',
              'shrink_share')
    def test_delegated_methods(self, method):
        conf, _layout = self._setup()
        _driver = GlusterfsFakeShareDriver(False, configuration=conf)
        fake_args = (mock.Mock(), mock.Mock(), mock.Mock())

        getattr(_driver, method)(*fake_args)

        getattr(_layout, method).assert_called_once_with(*fake_args)


@ddt.ddt
class GlusterfsShareLayoutBaseTestCase(test.TestCase):
    """Tests GlusterfsShareLayoutBaseTestCase."""

    def setUp(self):
        super(GlusterfsShareLayoutBaseTestCase, self).setUp()
        fake_utils.stub_out_utils_execute(self)
        self._execute = fake_utils.fake_execute
        self.addCleanup(fake_utils.fake_execute_set_repliers, [])
        self.addCleanup(fake_utils.fake_execute_clear_log)
        self.fake_driver = mock.Mock()
        self.mock_object(self.fake_driver, '_execute',
                         self._execute)

    class FakeLayout(layout.GlusterfsShareLayoutBase):

        def _share_manager(self, share):
            """Return GlusterManager object representing share's backend."""

        def do_setup(self, context):
            """Any initialization the share driver does while starting."""

        def create_share(self, context, share, share_server=None):
            """Is called to create share."""

        def create_share_from_snapshot(self, context, share, snapshot,
                                       share_server=None):
            """Is called to create share from snapshot."""

        def create_snapshot(self, context, snapshot, share_server=None):
            """Is called to create snapshot."""

        def delete_share(self, context, share, share_server=None):
            """Is called to remove share."""

        def delete_snapshot(self, context, snapshot, share_server=None):
            """Is called to remove snapshot."""

        def ensure_share(self, context, share, share_server=None):
            """Invoked to ensure that share is exported."""

        def manage_existing(self, share, driver_options):
            """Brings an existing share under Manila management."""

        def unmanage(self, share):
            """Removes the specified share from Manila management."""

        def extend_share(self, share, new_size, share_server=None):
            """Extends size of existing share."""

        def shrink_share(self, share, new_size, share_server=None):
            """Shrinks size of existing share."""

    def test_init_invalid(self):
        self.assertRaises(TypeError, layout.GlusterfsShareLayoutBase,
                          mock.Mock())

    def test_subclass(self):
        fake_conf = mock.Mock()
        _layout = self.FakeLayout(self.fake_driver, configuration=fake_conf)

        self.assertEqual(fake_conf, _layout.configuration)
        self.assertRaises(NotImplementedError, _layout._update_share_stats)

    def test_check_mount_glusterfs(self):
        fake_conf = mock.Mock()
        _driver = mock.Mock()
        _driver._execute = mock.Mock()
        _layout = self.FakeLayout(_driver, configuration=fake_conf)

        _layout._check_mount_glusterfs()

        _driver._execute.assert_called_once_with(
            'mount.glusterfs',
            check_exit_code=False)

    @ddt.data({'_errno': errno.ENOENT,
               '_exception': exception.GlusterfsException},
              {'_errno': errno.EACCES, '_exception': OSError})
    @ddt.unpack
    def test_check_mount_glusterfs_not_installed(self, _errno, _exception):
        fake_conf = mock.Mock()
        _layout = self.FakeLayout(self.fake_driver, configuration=fake_conf)

        def exec_runner(*ignore_args, **ignore_kwargs):
            raise OSError(_errno, os.strerror(_errno))

        expected_exec = ['mount.glusterfs']
        fake_utils.fake_execute_set_repliers([(expected_exec[0], exec_runner)])

        self.assertRaises(_exception, _layout._check_mount_glusterfs)



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For Simple Scheduler
"""

import mock
from oslo_config import cfg

from manila import context
from manila import db
from manila import exception
from manila.scheduler.drivers import base
from manila.scheduler.drivers import simple
from manila.share import rpcapi as share_rpcapi
from manila import test
from manila.tests import db_utils
from manila import utils

CONF = cfg.CONF


class SimpleSchedulerSharesTestCase(test.TestCase):
    """Test case for simple scheduler create share method."""

    def setUp(self):
        super(SimpleSchedulerSharesTestCase, self).setUp()
        self.mock_object(share_rpcapi, 'ShareAPI')
        self.driver = simple.SimpleScheduler()

        self.context = context.RequestContext('fake_user', 'fake_project')
        self.admin_context = context.RequestContext('fake_admin_user',
                                                    'fake_project')
        self.admin_context.is_admin = True

    @mock.patch.object(utils, 'service_is_up', mock.Mock(return_value=True))
    def test_create_share_if_two_services_up(self):
        share_id = 'fake'
        fake_share = {'id': share_id, 'size': 1}
        fake_service_1 = {'disabled': False, 'host': 'fake_host1'}
        fake_service_2 = {'disabled': False, 'host': 'fake_host2'}
        fake_result = [(fake_service_1, 2), (fake_service_2, 1)]
        fake_request_spec = {
            'share_id': share_id,
            'share_properties': fake_share,
        }
        self.mock_object(db, 'service_get_all_share_sorted',
                         mock.Mock(return_value=fake_result))
        self.mock_object(base, 'share_update_db',
                         mock.Mock(return_value=db_utils.create_share()))

        self.driver.schedule_create_share(self.context,
                                          fake_request_spec, {})
        utils.service_is_up.assert_called_once_with(utils.IsAMatcher(dict))
        db.service_get_all_share_sorted.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext))
        base.share_update_db.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext), share_id, 'fake_host1')

    def test_create_share_if_services_not_available(self):
        share_id = 'fake'
        fake_share = {'id': share_id, 'size': 1}
        fake_result = []
        fake_request_spec = {
            'share_id': share_id,
            'share_properties': fake_share,
        }
        with mock.patch.object(db, 'service_get_all_share_sorted',
                               mock.Mock(return_value=fake_result)):
            self.assertRaises(exception.NoValidHost,
                              self.driver.schedule_create_share,
                              self.context, fake_request_spec, {})
            db.service_get_all_share_sorted.assert_called_once_with(
                utils.IsAMatcher(context.RequestContext))

    def test_create_share_if_max_gigabytes_exceeded(self):
        share_id = 'fake'
        fake_share = {'id': share_id, 'size': 10001}
        fake_service_1 = {'disabled': False, 'host': 'fake_host1'}
        fake_service_2 = {'disabled': False, 'host': 'fake_host2'}
        fake_result = [(fake_service_1, 5), (fake_service_2, 7)]
        fake_request_spec = {
            'share_id': share_id,
            'share_properties': fake_share,
        }
        with mock.patch.object(db, 'service_get_all_share_sorted',
                               mock.Mock(return_value=fake_result)):
            self.assertRaises(exception.NoValidHost,
                              self.driver.schedule_create_share,
                              self.context, fake_request_spec, {})
            db.service_get_all_share_sorted.assert_called_once_with(
                utils.IsAMatcher(context.RequestContext))

    @mock.patch.object(utils, 'service_is_up', mock.Mock(return_value=True))
    def test_create_share_availability_zone(self):
        share_id = 'fake'
        fake_share = {
            'id': share_id,
            'size': 1,
        }
        fake_instance = {
            'availability_zone_id': 'fake',
        }
        fake_service_1 = {
            'disabled': False, 'host': 'fake_host1',
            'availability_zone_id': 'fake',
        }
        fake_service_2 = {
            'disabled': False, 'host': 'fake_host2',
            'availability_zone_id': 'super_fake',
        }
        fake_result = [(fake_service_1, 0), (fake_service_2, 1)]
        fake_request_spec = {
            'share_id': share_id,
            'share_properties': fake_share,
            'share_instance_properties': fake_instance,
        }
        self.mock_object(db, 'service_get_all_share_sorted',
                         mock.Mock(return_value=fake_result))
        self.mock_object(base, 'share_update_db',
                         mock.Mock(return_value=db_utils.create_share()))

        self.driver.schedule_create_share(self.context,
                                          fake_request_spec, {})
        utils.service_is_up.assert_called_once_with(fake_service_1)
        base.share_update_db.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext), share_id,
            fake_service_1['host'])
        db.service_get_all_share_sorted.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext))

    @mock.patch.object(utils, 'service_is_up', mock.Mock(return_value=True))
    def test_create_share_availability_zone_on_host(self):
        share_id = 'fake'
        fake_share = {
            'id': share_id,
            'availability_zone': 'fake:fake',
            'size': 1,
        }
        fake_service = {'disabled': False, 'host': 'fake'}
        fake_request_spec = {
            'share_id': share_id,
            'share_properties': fake_share,
        }
        self.mock_object(db, 'service_get_all_share_sorted',
                         mock.Mock(return_value=[(fake_service, 1)]))
        self.mock_object(base, 'share_update_db',
                         mock.Mock(return_value=db_utils.create_share()))

        self.driver.schedule_create_share(self.admin_context,
                                          fake_request_spec, {})
        utils.service_is_up.assert_called_once_with(fake_service)
        db.service_get_all_share_sorted.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext))
        base.share_update_db.assert_called_once_with(
            utils.IsAMatcher(context.RequestContext), share_id, 'fake')



# Copyright (c) 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import ddt
import mock
from oslo_utils import units


from manila.common import constants
from manila import context
import manila.exception as exception
from manila.share import configuration
from manila.share.drivers.cephfs import driver
from manila.share import share_types
from manila import test
from manila.tests import fake_share


DEFAULT_VOLUME_MODE = 0o755
ALT_VOLUME_MODE_CFG = '775'
ALT_VOLUME_MODE = 0o775


class MockVolumeClientModule(object):
    """Mocked up version of ceph's VolumeClient interface."""

    class VolumePath(object):
        """Copy of VolumePath from CephFSVolumeClient."""

        def __init__(self, group_id, volume_id):
            self.group_id = group_id
            self.volume_id = volume_id

        def __eq__(self, other):
            return (self.group_id == other.group_id
                    and self.volume_id == other.volume_id)

        def __str__(self):
            return "{0}/{1}".format(self.group_id, self.volume_id)

    class CephFSVolumeClient(mock.Mock):
        mock_used_bytes = 0
        version = 1

        def __init__(self, *args, **kwargs):
            mock.Mock.__init__(self, spec=[
                "connect", "disconnect",
                "create_snapshot_volume", "destroy_snapshot_volume",
                "create_group", "destroy_group",
                "delete_volume", "purge_volume",
                "deauthorize", "evict", "set_max_bytes",
                "destroy_snapshot_group", "create_snapshot_group",
                "get_authorized_ids"
            ])
            self.create_volume = mock.Mock(return_value={
                "mount_path": "/foo/bar"
            })
            self._get_path = mock.Mock(return_value='/foo/bar')
            self.get_mon_addrs = mock.Mock(return_value=["1.2.3.4", "5.6.7.8"])
            self.get_authorized_ids = mock.Mock(
                return_value=[('eve', 'rw')])
            self.authorize = mock.Mock(return_value={"auth_key": "abc123"})
            self.get_used_bytes = mock.Mock(return_value=self.mock_used_bytes)
            self.rados = mock.Mock()
            self.rados.get_cluster_stats = mock.Mock(return_value={
                "kb": 1000,
                "kb_avail": 500
            })


@ddt.ddt
class CephFSDriverTestCase(test.TestCase):
    """Test the CephFS driver.

    This is a very simple driver that mainly
    calls through to the CephFSVolumeClient interface, so the tests validate
    that the Manila driver calls map to the appropriate CephFSVolumeClient
    calls.
    """

    def setUp(self):
        super(CephFSDriverTestCase, self).setUp()
        self._execute = mock.Mock()
        self.fake_conf = configuration.Configuration(None)
        self._context = context.get_admin_context()
        self._share = fake_share.fake_share(share_proto='CEPHFS')

        self.fake_conf.set_default('driver_handles_share_servers', False)
        self.fake_conf.set_default('cephfs_auth_id', 'manila')

        self.mock_object(driver, "ceph_volume_client",
                         MockVolumeClientModule)
        self.mock_object(driver, "ceph_module_found", True)
        self.mock_object(driver, "cephfs_share_path")
        self.mock_object(driver, 'NativeProtocolHelper')
        self.mock_object(driver, 'NFSProtocolHelper')

        self._driver = (
            driver.CephFSDriver(execute=self._execute,
                                configuration=self.fake_conf))
        self._driver.protocol_helper = mock.Mock()

        self.mock_object(share_types, 'get_share_type_extra_specs',
                         mock.Mock(return_value={}))

    @ddt.data('cephfs', 'nfs')
    def test_do_setup(self, protocol_helper):
        self._driver.configuration.cephfs_protocol_helper_type = (
            protocol_helper)

        self._driver.do_setup(self._context)

        if protocol_helper == 'cephfs':
            driver.NativeProtocolHelper.assert_called_once_with(
                self._execute, self._driver.configuration,
                ceph_vol_client=self._driver._volume_client)
        else:
            driver.NFSProtocolHelper.assert_called_once_with(
                self._execute, self._driver.configuration,
                ceph_vol_client=self._driver._volume_client)

        self._driver.protocol_helper.init_helper.assert_called_once_with()

        self.assertEqual(DEFAULT_VOLUME_MODE, self._driver._cephfs_volume_mode)

    def test_create_share(self):
        cephfs_volume = {"mount_path": "/foo/bar"}

        self._driver.create_share(self._context, self._share)

        self._driver._volume_client.create_volume.assert_called_once_with(
            driver.cephfs_share_path(self._share),
            size=self._share['size'] * units.Gi,
            data_isolated=False, mode=DEFAULT_VOLUME_MODE)
        (self._driver.protocol_helper.get_export_locations.
            assert_called_once_with(self._share, cephfs_volume))

    def test_create_share_error(self):
        share = fake_share.fake_share(share_proto='NFS')

        self.assertRaises(exception.ShareBackendException,
                          self._driver.create_share,
                          self._context,
                          share)

    def test_update_access(self):
        alice = {
            'id': 'instance_mapping_id1',
            'access_id': 'accessid1',
            'access_level': 'rw',
            'access_type': 'cephx',
            'access_to': 'alice'
        }
        add_rules = access_rules = [alice, ]
        delete_rules = []

        self._driver.update_access(
            self._context, self._share, access_rules, add_rules, delete_rules,
            None)

        self._driver.protocol_helper.update_access.assert_called_once_with(
            self._context, self._share, access_rules, add_rules, delete_rules,
            share_server=None)

    def test_ensure_share(self):
        self._driver.ensure_share(self._context,
                                  self._share)

        self._driver._volume_client.create_volume.assert_called_once_with(
            driver.cephfs_share_path(self._share),
            size=self._share['size'] * units.Gi,
            data_isolated=False,
            mode=DEFAULT_VOLUME_MODE)

    def test_create_data_isolated(self):
        self.mock_object(share_types, 'get_share_type_extra_specs',
                         mock.Mock(return_value={"cephfs:data_isolated": True})
                         )

        self._driver.create_share(self._context, self._share)

        self._driver._volume_client.create_volume.assert_called_once_with(
            driver.cephfs_share_path(self._share),
            size=self._share['size'] * units.Gi,
            data_isolated=True,
            mode=DEFAULT_VOLUME_MODE)

    def test_delete_share(self):
        self._driver.delete_share(self._context, self._share)

        self._driver._volume_client.delete_volume.assert_called_once_with(
            driver.cephfs_share_path(self._share),
            data_isolated=False)
        self._driver._volume_client.purge_volume.assert_called_once_with(
            driver.cephfs_share_path(self._share),
            data_isolated=False)

    def test_delete_data_isolated(self):
        self.mock_object(share_types, 'get_share_type_extra_specs',
                         mock.Mock(return_value={"cephfs:data_isolated": True})
                         )

        self._driver.delete_share(self._context, self._share)

        self._driver._volume_client.delete_volume.assert_called_once_with(
            driver.cephfs_share_path(self._share),
            data_isolated=True)
        self._driver._volume_client.purge_volume.assert_called_once_with(
            driver.cephfs_share_path(self._share),
            data_isolated=True)

    def test_extend_share(self):
        new_size_gb = self._share['size'] * 2
        new_size = new_size_gb * units.Gi

        self._driver.extend_share(self._share, new_size_gb, None)

        self._driver._volume_client.set_max_bytes.assert_called_once_with(
            driver.cephfs_share_path(self._share),
            new_size)

    def test_shrink_share(self):
        new_size_gb = self._share['size'] * 0.5
        new_size = new_size_gb * units.Gi

        self._driver.shrink_share(self._share, new_size_gb, None)

        self._driver._volume_client.get_used_bytes.assert_called_once_with(
            driver.cephfs_share_path(self._share))
        self._driver._volume_client.set_max_bytes.assert_called_once_with(
            driver.cephfs_share_path(self._share),
            new_size)

    def test_shrink_share_full(self):
        """That shrink fails when share is too full."""
        new_size_gb = self._share['size'] * 0.5

        # Pretend to be full up
        vc = MockVolumeClientModule.CephFSVolumeClient
        vc.mock_used_bytes = (units.Gi * self._share['size'])

        self.assertRaises(exception.ShareShrinkingPossibleDataLoss,
                          self._driver.shrink_share,
                          self._share, new_size_gb, None)
        self._driver._volume_client.set_max_bytes.assert_not_called()

    def test_create_snapshot(self):
        self._driver.create_snapshot(self._context,
                                     {
                                         "id": "instance1",
                                         "share": self._share,
                                         "snapshot_id": "snappy1"
                                     },
                                     None)

        (self._driver._volume_client.create_snapshot_volume
            .assert_called_once_with(
                driver.cephfs_share_path(self._share),
                "snappy1_instance1",
                mode=DEFAULT_VOLUME_MODE))

    def test_delete_snapshot(self):
        self._driver.delete_snapshot(self._context,
                                     {
                                         "id": "instance1",
                                         "share": self._share,
                                         "snapshot_id": "snappy1"
                                     },
                                     None)

        (self._driver._volume_client.destroy_snapshot_volume
            .assert_called_once_with(
                driver.cephfs_share_path(self._share),
                "snappy1_instance1"))

    def test_create_share_group(self):
        self._driver.create_share_group(self._context, {"id": "grp1"}, None)

        self._driver._volume_client.create_group.assert_called_once_with(
            "grp1", mode=DEFAULT_VOLUME_MODE)

    def test_delete_share_group(self):
        self._driver.delete_share_group(self._context, {"id": "grp1"}, None)

        self._driver._volume_client.destroy_group.assert_called_once_with(
            "grp1")

    def test_create_share_snapshot(self):
        self._driver.create_share_group_snapshot(self._context, {
            'share_group_id': 'sgid',
            'id': 'snapid',
        })

        (self._driver._volume_client.create_snapshot_group.
         assert_called_once_with("sgid", "snapid", mode=DEFAULT_VOLUME_MODE))

    def test_delete_share_group_snapshot(self):
        self._driver.delete_share_group_snapshot(self._context, {
            'share_group_id': 'sgid',
            'id': 'snapid',
        })

        (self._driver._volume_client.destroy_snapshot_group.
         assert_called_once_with("sgid", "snapid"))

    def test_delete_driver(self):
        # Create share to prompt volume_client construction
        self._driver.create_share(self._context,
                                  self._share)

        vc = self._driver._volume_client
        del self._driver

        vc.disconnect.assert_called_once_with()

    def test_delete_driver_no_client(self):
        self.assertIsNone(self._driver._volume_client)
        del self._driver

    def test_connect_noevict(self):
        # When acting as "admin", driver should skip evicting
        self._driver.configuration.local_conf.set_override('cephfs_auth_id',
                                                           "admin")

        self._driver.create_share(self._context,
                                  self._share)

        vc = self._driver._volume_client
        vc.connect.assert_called_once_with(premount_evict=None)

    def test_update_share_stats(self):
        self._driver._volume_client
        self._driver._update_share_stats()
        result = self._driver._stats

        self.assertTrue(result['ipv4_support'])
        self.assertFalse(result['ipv6_support'])
        self.assertEqual("CEPHFS", result['storage_protocol'])

    def test_module_missing(self):
        driver.ceph_module_found = False
        driver.ceph_volume_client = None

        self.assertRaises(exception.ManilaException,
                          self._driver.create_share,
                          self._context,
                          self._share)


@ddt.ddt
class NativeProtocolHelperTestCase(test.TestCase):

    def setUp(self):
        super(NativeProtocolHelperTestCase, self).setUp()
        self.fake_conf = configuration.Configuration(None)
        self._context = context.get_admin_context()
        self._share = fake_share.fake_share(share_proto='CEPHFS')

        self.fake_conf.set_default('driver_handles_share_servers', False)

        self.mock_object(driver, "cephfs_share_path")

        self._native_protocol_helper = driver.NativeProtocolHelper(
            None,
            self.fake_conf,
            ceph_vol_client=MockVolumeClientModule.CephFSVolumeClient()
        )

    def test_get_export_locations(self):
        vc = self._native_protocol_helper.volume_client
        fake_cephfs_volume = {'mount_path': '/foo/bar'}
        expected_export_locations = {
            'path': '1.2.3.4,5.6.7.8:/foo/bar',
            'is_admin_only': False,
            'metadata': {},
        }

        export_locations = self._native_protocol_helper.get_export_locations(
            self._share, fake_cephfs_volume)

        self.assertEqual(expected_export_locations, export_locations)
        vc.get_mon_addrs.assert_called_once_with()

    @ddt.data(None, 1)
    def test_allow_access_rw(self, volume_client_version):
        vc = self._native_protocol_helper.volume_client
        rule = {
            'access_level': constants.ACCESS_LEVEL_RW,
            'access_to': 'alice',
            'access_type': 'cephx',
        }
        vc.version = volume_client_version

        auth_key = self._native_protocol_helper._allow_access(
            self._context, self._share, rule)

        self.assertEqual("abc123", auth_key)

        if not volume_client_version:
            vc.authorize.assert_called_once_with(
                driver.cephfs_share_path(self._share), "alice")
        else:
            vc.authorize.assert_called_once_with(
                driver.cephfs_share_path(self._share), "alice",
                readonly=False, tenant_id=self._share['project_id'])

    @ddt.data(None, 1)
    def test_allow_access_ro(self, volume_client_version):
        vc = self._native_protocol_helper.volume_client
        rule = {
            'access_level': constants.ACCESS_LEVEL_RO,
            'access_to': 'alice',
            'access_type': 'cephx',
        }
        vc.version = volume_client_version

        if not volume_client_version:
            self.assertRaises(exception.InvalidShareAccessLevel,
                              self._native_protocol_helper._allow_access,
                              self._context, self._share, rule)
        else:
            auth_key = (
                self._native_protocol_helper._allow_access(
                    self._context, self._share, rule)
            )

            self.assertEqual("abc123", auth_key)
            vc.authorize.assert_called_once_with(
                driver.cephfs_share_path(self._share), "alice", readonly=True,
                tenant_id=self._share['project_id'])

    def test_allow_access_wrong_type(self):
        self.assertRaises(exception.InvalidShareAccess,
                          self._native_protocol_helper._allow_access,
                          self._context, self._share, {
                              'access_level': constants.ACCESS_LEVEL_RW,
                              'access_type': 'RHUBARB',
                              'access_to': 'alice'
                          })

    def test_allow_access_same_cephx_id_as_manila_service(self):
        self.assertRaises(exception.InvalidInput,
                          self._native_protocol_helper._allow_access,
                          self._context, self._share, {
                              'access_level': constants.ACCESS_LEVEL_RW,
                              'access_type': 'cephx',
                              'access_to': 'manila',
                          })

    def test_deny_access(self):
        vc = self._native_protocol_helper.volume_client
        self._native_protocol_helper._deny_access(self._context, self._share, {
            'access_level': 'rw',
            'access_type': 'cephx',
            'access_to': 'alice'
        })

        vc.deauthorize.assert_called_once_with(
            driver.cephfs_share_path(self._share), "alice")
        vc.evict.assert_called_once_with(
            "alice", volume_path=driver.cephfs_share_path(self._share))

    def test_update_access_add_rm(self):
        vc = self._native_protocol_helper.volume_client
        alice = {
            'id': 'instance_mapping_id1',
            'access_id': 'accessid1',
            'access_level': 'rw',
            'access_type': 'cephx',
            'access_to': 'alice'
        }
        bob = {
            'id': 'instance_mapping_id2',
            'access_id': 'accessid2',
            'access_level': 'rw',
            'access_type': 'cephx',
            'access_to': 'bob'
        }

        access_updates = self._native_protocol_helper.update_access(
            self._context, self._share, access_rules=[alice],
            add_rules=[alice], delete_rules=[bob])

        self.assertEqual(
            {'accessid1': {'access_key': 'abc123'}}, access_updates)
        vc.authorize.assert_called_once_with(
            driver.cephfs_share_path(self._share), "alice", readonly=False,
            tenant_id=self._share['project_id'])
        vc.deauthorize.assert_called_once_with(
            driver.cephfs_share_path(self._share), "bob")

    @ddt.data(None, 1)
    def test_update_access_all(self, volume_client_version):
        vc = self._native_protocol_helper.volume_client
        alice = {
            'id': 'instance_mapping_id1',
            'access_id': 'accessid1',
            'access_level': 'rw',
            'access_type': 'cephx',
            'access_to': 'alice'
        }
        vc.version = volume_client_version

        access_updates = self._native_protocol_helper.update_access(
            self._context, self._share, access_rules=[alice], add_rules=[],
            delete_rules=[])

        self.assertEqual(
            {'accessid1': {'access_key': 'abc123'}}, access_updates)

        if volume_client_version:
            vc.get_authorized_ids.assert_called_once_with(
                driver.cephfs_share_path(self._share))
            vc.authorize.assert_called_once_with(
                driver.cephfs_share_path(self._share), "alice", readonly=False,
                tenant_id=self._share['project_id'])
            vc.deauthorize.assert_called_once_with(
                driver.cephfs_share_path(self._share), "eve")
        else:
            self.assertFalse(vc.get_authorized_ids.called)
            vc.authorize.assert_called_once_with(
                driver.cephfs_share_path(self._share), "alice")


@ddt.ddt
class NFSProtocolHelperTestCase(test.TestCase):

    def setUp(self):
        super(NFSProtocolHelperTestCase, self).setUp()
        self._execute = mock.Mock()
        self._share = fake_share.fake_share(share_proto='NFS')
        self._volume_client = MockVolumeClientModule.CephFSVolumeClient()
        self.fake_conf = configuration.Configuration(None)

        self.fake_conf.set_default('cephfs_ganesha_server_ip',
                                   'fakeip')
        self.mock_object(driver, "cephfs_share_path",
                         mock.Mock(return_value='fakevolumepath'))
        self.mock_object(driver.ganesha_utils, 'SSHExecutor')
        self.mock_object(driver.ganesha_utils, 'RootExecutor')
        self.mock_object(driver.socket, 'gethostname')

        self._nfs_helper = driver.NFSProtocolHelper(
            self._execute,
            self.fake_conf,
            ceph_vol_client=self._volume_client)

    @ddt.data(False, True)
    def test_init_executor_type(self, ganesha_server_is_remote):
        fake_conf = configuration.Configuration(None)
        conf_args_list = [
            ('cephfs_ganesha_server_is_remote', ganesha_server_is_remote),
            ('cephfs_ganesha_server_ip', 'fakeip'),
            ('cephfs_ganesha_server_username', 'fake_username'),
            ('cephfs_ganesha_server_password', 'fakepwd'),
            ('cephfs_ganesha_path_to_private_key', 'fakepathtokey')]
        for args in conf_args_list:
            fake_conf.set_default(*args)

        driver.NFSProtocolHelper(
            self._execute,
            fake_conf,
            ceph_vol_client=MockVolumeClientModule.CephFSVolumeClient()
        )

        if ganesha_server_is_remote:
            driver.ganesha_utils.SSHExecutor.assert_has_calls(
                [mock.call('fakeip', 22, None, 'fake_username',
                           password='fakepwd',
                           privatekey='fakepathtokey')])
        else:
            driver.ganesha_utils.RootExecutor.assert_has_calls(
                [mock.call(self._execute)])

    @ddt.data('fakeip', None)
    def test_init_identify_local_host(self, ganesha_server_ip):
        self.mock_object(driver.LOG, 'info')
        fake_conf = configuration.Configuration(None)
        conf_args_list = [
            ('cephfs_ganesha_server_ip', ganesha_server_ip),
            ('cephfs_ganesha_server_username', 'fake_username'),
            ('cephfs_ganesha_server_password', 'fakepwd'),
            ('cephfs_ganesha_path_to_private_key', 'fakepathtokey')]
        for args in conf_args_list:
            fake_conf.set_default(*args)

        driver.NFSProtocolHelper(
            self._execute,
            fake_conf,
            ceph_vol_client=MockVolumeClientModule.CephFSVolumeClient()
        )

        driver.ganesha_utils.RootExecutor.assert_has_calls(
            [mock.call(self._execute)])
        if ganesha_server_ip:
            self.assertFalse(driver.socket.gethostname.called)
            self.assertFalse(driver.LOG.info.called)
        else:
            driver.socket.gethostname.assert_called_once_with()
            driver.LOG.info.assert_called_once()

    def test_get_export_locations(self):
        cephfs_volume = {"mount_path": "/foo/bar"}

        ret = self._nfs_helper.get_export_locations(self._share,
                                                    cephfs_volume)
        self.assertEqual(
            {
                'path': 'fakeip:/foo/bar',
                'is_admin_only': False,
                'metadata': {}
            }, ret)

    def test_default_config_hook(self):
        fake_conf_dict = {'key': 'value1'}
        self.mock_object(driver.ganesha.GaneshaNASHelper,
                         '_default_config_hook',
                         mock.Mock(return_value={}))
        self.mock_object(driver.ganesha_utils, 'path_from',
                         mock.Mock(return_value='/fakedir/cephfs/conf'))
        self.mock_object(self._nfs_helper, '_load_conf_dir',
                         mock.Mock(return_value=fake_conf_dict))

        ret = self._nfs_helper._default_config_hook()

        (driver.ganesha.GaneshaNASHelper._default_config_hook.
            assert_called_once_with())
        driver.ganesha_utils.path_from.assert_called_once_with(
            driver.__file__, 'conf')
        self._nfs_helper._load_conf_dir.assert_called_once_with(
            '/fakedir/cephfs/conf')
        self.assertEqual(fake_conf_dict, ret)

    def test_fsal_hook(self):
        expected_ret = {
            'Name': 'Ceph',
            'User_Id': 'ganesha-fakeid',
            'Secret_Access_Key': 'fakekey'
        }
        self.mock_object(self._volume_client, 'authorize',
                         mock.Mock(return_value={'auth_key': 'fakekey'}))

        ret = self._nfs_helper._fsal_hook(None, self._share, None)

        driver.cephfs_share_path.assert_called_once_with(self._share)
        self._volume_client.authorize.assert_called_once_with(
            'fakevolumepath', 'ganesha-fakeid', readonly=False,
            tenant_id='fake_project_uuid')
        self.assertEqual(expected_ret, ret)

    def test_cleanup_fsal_hook(self):
        self.mock_object(self._volume_client, 'deauthorize')

        ret = self._nfs_helper._cleanup_fsal_hook(None, self._share, None)

        driver.cephfs_share_path.assert_called_once_with(self._share)
        self._volume_client.deauthorize.assert_called_once_with(
            'fakevolumepath', 'ganesha-fakeid')
        self.assertIsNone(ret)

    def test_get_export_path(self):
        ret = self._nfs_helper._get_export_path(self._share)

        driver.cephfs_share_path.assert_called_once_with(self._share)
        self._volume_client._get_path.assert_called_once_with(
            'fakevolumepath')
        self.assertEqual('/foo/bar', ret)

    def test_get_export_pseudo_path(self):
        ret = self._nfs_helper._get_export_pseudo_path(self._share)

        driver.cephfs_share_path.assert_called_once_with(self._share)
        self._volume_client._get_path.assert_called_once_with(
            'fakevolumepath')
        self.assertEqual('/foo/bar', ret)


@ddt.ddt
class CephFSDriverAltConfigTestCase(test.TestCase):
    """Test the CephFS driver with non-default config values."""

    def setUp(self):
        super(CephFSDriverAltConfigTestCase, self).setUp()
        self._execute = mock.Mock()
        self.fake_conf = configuration.Configuration(None)
        self._context = context.get_admin_context()
        self._share = fake_share.fake_share(share_proto='CEPHFS')

        self.fake_conf.set_default('driver_handles_share_servers', False)
        self.fake_conf.set_default('cephfs_auth_id', 'manila')

        self.mock_object(driver, "ceph_volume_client",
                         MockVolumeClientModule)
        self.mock_object(driver, "ceph_module_found", True)
        self.mock_object(driver, "cephfs_share_path")
        self.mock_object(driver, 'NativeProtocolHelper')
        self.mock_object(driver, 'NFSProtocolHelper')

    @ddt.data('cephfs', 'nfs')
    def test_do_setup_alt_volume_mode(self, protocol_helper):

        self.fake_conf.set_default('cephfs_volume_mode', ALT_VOLUME_MODE_CFG)
        self._driver = driver.CephFSDriver(execute=self._execute,
                                           configuration=self.fake_conf)

        self._driver.configuration.cephfs_protocol_helper_type = (
            protocol_helper)

        self._driver.do_setup(self._context)

        if protocol_helper == 'cephfs':
            driver.NativeProtocolHelper.assert_called_once_with(
                self._execute, self._driver.configuration,
                ceph_vol_client=self._driver._volume_client)
        else:
            driver.NFSProtocolHelper.assert_called_once_with(
                self._execute, self._driver.configuration,
                ceph_vol_client=self._driver._volume_client)

        self._driver.protocol_helper.init_helper.assert_called_once_with()

        self.assertEqual(ALT_VOLUME_MODE, self._driver._cephfs_volume_mode)

    @ddt.data('0o759', '0x755', '12a3')
    def test_volume_mode_exception(self, volume_mode):
        # cephfs_volume_mode must be a string representing an int as octal
        self.fake_conf.set_default('cephfs_volume_mode', volume_mode)

        self.assertRaises(exception.BadConfigurationException,
                          driver.CephFSDriver, execute=self._execute,
                          configuration=self.fake_conf)



# Copyright (c) 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock

import os

from manila.share import configuration
from manila.share.drivers import generic
from manila.share.drivers.windows import service_instance
from manila.share.drivers.windows import windows_smb_driver as windows_drv
from manila.share.drivers.windows import windows_smb_helper
from manila.share.drivers.windows import windows_utils
from manila.share.drivers.windows import winrm_helper
from manila import test
from manila.tests import fake_share


@ddt.ddt
class WindowsSMBDriverTestCase(test.TestCase):
    @mock.patch.object(winrm_helper, 'WinRMHelper')
    @mock.patch.object(windows_utils, 'WindowsUtils')
    @mock.patch.object(windows_smb_helper, 'WindowsSMBHelper')
    @mock.patch.object(service_instance,
                       'WindowsServiceInstanceManager')
    def setUp(self, mock_sv_instance_mgr, mock_smb_helper_cls,
              mock_utils_cls, mock_winrm_helper_cls):
        self.flags(driver_handles_share_servers=True)
        self._fake_conf = configuration.Configuration(None)

        self._share = fake_share.fake_share(share_proto='SMB')
        self._share_server = dict(
            backend_details=mock.sentinel.backend_details)

        self._drv = windows_drv.WindowsSMBDriver(
            configuration=self._fake_conf)
        self._drv._setup_helpers()

        self._remote_execute = mock_winrm_helper_cls.return_value
        self._windows_utils = mock_utils_cls.return_value
        self._smb_helper = mock_smb_helper_cls.return_value
        super(WindowsSMBDriverTestCase, self).setUp()

    @mock.patch('manila.share.driver.ShareDriver')
    def test_update_share_stats(self, mock_base_driver):
        self._drv._update_share_stats()
        mock_base_driver._update_share_stats.assert_called_once_with(
            self._drv,
            data=dict(storage_protocol="CIFS"))

    @mock.patch.object(service_instance, 'WindowsServiceInstanceManager')
    def test_setup_service_instance_manager(self, mock_sv_instance_mgr):
        self._drv._setup_service_instance_manager()
        mock_sv_instance_mgr.assert_called_once_with(
            driver_config=self._fake_conf)

    def test_setup_helpers(self):
        expected_helpers = {"SMB": self._smb_helper,
                            "CIFS": self._smb_helper}
        self._drv._setup_helpers()
        self.assertEqual(expected_helpers, self._drv._helpers)

    @mock.patch.object(generic.GenericShareDriver, '_teardown_server')
    def test_teardown_server(self, mock_super_teardown):
        mock_server = {'joined_domain': True,
                       'instance_id': mock.sentinel.instance_id}
        mock_sec_service = {'user': mock.sentinel.user,
                            'password': mock.sentinel.password,
                            'domain': mock.sentinel.domain}

        sv_mgr = self._drv.service_instance_manager
        sv_mgr.get_valid_security_service.return_value = mock_sec_service
        # We ensure that domain unjoin exceptions do not prevent the
        # service instance from being teared down.
        self._windows_utils.unjoin_domain.side_effect = Exception

        self._drv._teardown_server(mock_server,
                                   mock_sec_service)

        sv_mgr.get_valid_security_service.assert_called_once_with(
            mock_sec_service)
        self._windows_utils.unjoin_domain.assert_called_once_with(
            mock_server,
            mock_sec_service['user'],
            mock_sec_service['password'])
        mock_super_teardown.assert_called_once_with(mock_server,
                                                    mock_sec_service)

    @mock.patch.object(windows_drv.WindowsSMBDriver, '_get_disk_number')
    def test_format_device(self, mock_get_disk_number):
        mock_get_disk_number.return_value = mock.sentinel.disk_number

        self._drv._format_device(mock.sentinel.server, mock.sentinel.vol)

        self._drv._get_disk_number.assert_called_once_with(
            mock.sentinel.server, mock.sentinel.vol)
        self._windows_utils.initialize_disk.assert_called_once_with(
            mock.sentinel.server, mock.sentinel.disk_number)
        self._windows_utils.create_partition.assert_called_once_with(
            mock.sentinel.server, mock.sentinel.disk_number)
        self._windows_utils.format_partition.assert_called_once_with(
            mock.sentinel.server, mock.sentinel.disk_number,
            self._drv._DEFAULT_SHARE_PARTITION)

    @mock.patch.object(windows_drv.WindowsSMBDriver,
                       '_ensure_disk_online_and_writable')
    @mock.patch.object(windows_drv.WindowsSMBDriver, '_get_disk_number')
    @mock.patch.object(windows_drv.WindowsSMBDriver, '_get_mount_path')
    @mock.patch.object(windows_drv.WindowsSMBDriver, '_is_device_mounted')
    def test_mount_device(self, mock_device_mounted, mock_get_mount_path,
                          mock_get_disk_number, mock_ensure_disk):
        mock_get_mount_path.return_value = mock.sentinel.mount_path
        mock_get_disk_number.return_value = mock.sentinel.disk_number
        mock_device_mounted.return_value = False

        self._drv._mount_device(share=mock.sentinel.share,
                                server_details=mock.sentinel.server,
                                volume=mock.sentinel.vol)

        mock_device_mounted.assert_called_once_with(
            mock.sentinel.mount_path, mock.sentinel.server, mock.sentinel.vol)
        mock_get_disk_number.assert_called_once_with(
            mock.sentinel.server, mock.sentinel.vol)
        self._windows_utils.ensure_directory_exists.assert_called_once_with(
            mock.sentinel.server, mock.sentinel.mount_path)
        self._windows_utils.add_access_path(
            mock.sentinel.server,
            mock.sentinel.mount_path,
            mock.sentinel.disk_number,
            self._drv._DEFAULT_SHARE_PARTITION)
        mock_ensure_disk.assert_called_once_with(
            mock.sentinel.server, mock.sentinel.disk_number)

    @mock.patch.object(windows_drv.WindowsSMBDriver, '_get_mount_path')
    def test_unmount_device(self, mock_get_mount_path):
        mock_get_mount_path.return_value = mock.sentinel.mount_path
        mock_get_disk_number_by_path = (
            self._windows_utils.get_disk_number_by_mount_path)

        self._drv._unmount_device(mock.sentinel.share,
                                  mock.sentinel.server)

        mock_get_mount_path.assert_called_once_with(mock.sentinel.share)

        mock_get_disk_number_by_path.assert_called_once_with(
            mock.sentinel.server, mock.sentinel.mount_path)
        self._windows_utils.set_disk_online_status.assert_called_once_with(
            mock.sentinel.server,
            mock_get_disk_number_by_path.return_value,
            online=False)

    @ddt.data(None, 1)
    @mock.patch.object(windows_drv.WindowsSMBDriver, '_get_disk_number')
    @mock.patch.object(windows_drv.WindowsSMBDriver,
                       '_ensure_disk_online_and_writable')
    def test_resize_filesystem(self, new_size, mock_ensure_disk,
                               mock_get_disk_number):
        mock_get_disk_number.return_value = mock.sentinel.disk_number
        mock_get_max_size = self._windows_utils.get_partition_maximum_size
        mock_get_max_size.return_value = mock.sentinel.max_size

        self._drv._resize_filesystem(mock.sentinel.server,
                                     mock.sentinel.vol,
                                     new_size=new_size)

        mock_get_disk_number.assert_called_once_with(mock.sentinel.server,
                                                     mock.sentinel.vol)
        self._drv._ensure_disk_online_and_writable.assert_called_once_with(
            mock.sentinel.server, mock.sentinel.disk_number)

        if not new_size:
            mock_get_max_size.assert_called_once_with(
                mock.sentinel.server,
                mock.sentinel.disk_number,
                self._drv._DEFAULT_SHARE_PARTITION)
            expected_new_size = mock.sentinel.max_size
        else:
            expected_new_size = new_size << 30

        self._windows_utils.resize_partition.assert_called_once_with(
            mock.sentinel.server,
            expected_new_size,
            mock.sentinel.disk_number,
            self._drv._DEFAULT_SHARE_PARTITION)

    def test_ensure_disk_online_and_writable(self):
        self._drv._ensure_disk_online_and_writable(
            mock.sentinel.server, mock.sentinel.disk_number)

        self._windows_utils.update_disk.assert_called_once_with(
            mock.sentinel.server, mock.sentinel.disk_number)
        self._windows_utils.set_disk_online_status.assert_called_once_with(
            mock.sentinel.server, mock.sentinel.disk_number, online=True)
        self._windows_utils.set_disk_readonly_status.assert_called_once_with(
            mock.sentinel.server, mock.sentinel.disk_number, readonly=False)

    def test_get_mounted_share_size(self):
        fake_size_gb = 10
        self._windows_utils.get_disk_space_by_path.return_value = (
            fake_size_gb << 30, mock.sentinel.free_bytes)

        share_size = self._drv._get_mounted_share_size(
            mock.sentinel.mount_path,
            mock.sentinel.server)

        self.assertEqual(fake_size_gb, share_size)

    def test_get_consumed_space(self):
        fake_size_gb = 2
        fake_free_space_gb = 1
        self._windows_utils.get_disk_space_by_path.return_value = (
            fake_size_gb << 30, fake_free_space_gb << 30)

        consumed_space = self._drv._get_consumed_space(
            mock.sentinel.mount_path,
            mock.sentinel.server)

        self.assertEqual(fake_size_gb - fake_free_space_gb, consumed_space)

    def test_get_mount_path(self):
        fake_mount_path = 'fake_mount_path'
        fake_share_name = 'fake_share_name'
        mock_share = {'name': fake_share_name}
        self.flags(share_mount_path=fake_mount_path)

        mount_path = self._drv._get_mount_path(mock_share)

        self._windows_utils.normalize_path.assert_called_once_with(
            os.path.join(fake_mount_path, fake_share_name))
        self.assertEqual(self._windows_utils.normalize_path.return_value,
                         mount_path)

    @ddt.data(None, 2)
    def test_get_disk_number(self, disk_number_by_serial=None):
        mock_get_disk_number_by_serial = (
            self._windows_utils.get_disk_number_by_serial_number)

        mock_get_disk_number_by_serial.return_value = disk_number_by_serial
        mock_volume = {'id': mock.sentinel.vol_id,
                       'mountpoint': "/dev/sdb"}
        # If the disk number cannot be identified using the disk serial
        # number, we expect it to be retrieved based on the volume mountpoint,
        # having disk number 1 in this case.
        expected_disk_number = (disk_number_by_serial
                                if disk_number_by_serial else 1)

        disk_number = self._drv._get_disk_number(mock.sentinel.server,
                                                 mock_volume)

        mock_get_disk_number_by_serial.assert_called_once_with(
            mock.sentinel.server, mock.sentinel.vol_id)
        self.assertEqual(expected_disk_number, disk_number)

    @ddt.data(None, 2)
    def test_is_device_mounted(self, disk_number_by_path):
        mock_get_disk_number_by_path = (
            self._windows_utils.get_disk_number_by_mount_path)
        mock_get_disk_number_by_path.return_value = disk_number_by_path

        expected_result = disk_number_by_path is not None
        is_mounted = self._drv._is_device_mounted(
            mount_path=mock.sentinel.mount_path,
            server_details=mock.sentinel.server)

        mock_get_disk_number_by_path.assert_called_once_with(
            mock.sentinel.server,
            mock.sentinel.mount_path)
        self.assertEqual(expected_result, is_mounted)



# Copyright (c) 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from manila.scheduler.filters import base_host
from manila.scheduler import utils

LOG = log.getLogger(__name__)


class CapabilitiesFilter(base_host.BaseHostFilter):
    """HostFilter to work with resource (instance & volume) type records."""

    def _satisfies_extra_specs(self, capabilities, resource_type):
        """Compare capabilities against extra specs.

        Check that the capabilities provided by the services satisfy
        the extra specs associated with the resource type.
        """
        extra_specs = resource_type.get('extra_specs', [])
        if not extra_specs:
            return True

        return utils.capabilities_satisfied(capabilities, extra_specs)

    def host_passes(self, host_state, filter_properties):
        """Return a list of hosts that can create resource_type."""
        # Note(zhiteng) Currently only Cinder and Nova are using
        # this filter, so the resource type is either instance or
        # volume.
        resource_type = filter_properties.get('resource_type')
        if not self._satisfies_extra_specs(host_state.capabilities,
                                           resource_type):
            LOG.debug("%(host_state)s fails resource_type extra_specs "
                      "requirements", {'host_state': host_state})
            return False
        return True



# Copyright (c) 2014 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" GlusterFS native protocol (glusterfs) driver for shares.

Test cases for GlusterFS native protocol driver.
"""

import ddt
import mock
from oslo_config import cfg

from manila.common import constants
from manila import context
from manila import exception
from manila.share import configuration as config
from manila.share.drivers.glusterfs import common
from manila.share.drivers.glusterfs import glusterfs_native
from manila import test
from manila.tests import fake_utils


CONF = cfg.CONF


def new_share(**kwargs):
    share = {
        'id': 'fakeid',
        'name': 'fakename',
        'size': 1,
        'share_proto': 'glusterfs',
    }
    share.update(kwargs)
    return share


@ddt.ddt
class GlusterfsNativeShareDriverTestCase(test.TestCase):
    """Tests GlusterfsNativeShareDriver."""

    def setUp(self):
        super(GlusterfsNativeShareDriverTestCase, self).setUp()
        fake_utils.stub_out_utils_execute(self)
        self._execute = fake_utils.fake_execute
        self._context = context.get_admin_context()

        self.glusterfs_target1 = 'root@host1:/gv1'
        self.glusterfs_target2 = 'root@host2:/gv2'
        self.glusterfs_server1 = 'root@host1'
        self.glusterfs_server2 = 'root@host2'
        self.glusterfs_server1_volumes = 'manila-share-1-1G\nshare1'
        self.glusterfs_server2_volumes = 'manila-share-2-2G\nshare2'
        self.share1 = new_share(
            export_location=self.glusterfs_target1,
            status=constants.STATUS_AVAILABLE)
        self.share2 = new_share(
            export_location=self.glusterfs_target2,
            status=constants.STATUS_AVAILABLE)
        self.gmgr1 = common.GlusterManager(self.glusterfs_server1,
                                           self._execute, None, None,
                                           requires={'volume': False})
        self.gmgr2 = common.GlusterManager(self.glusterfs_server2,
                                           self._execute, None, None,
                                           requires={'volume': False})
        self.glusterfs_volumes_dict = (
            {'root@host1:/manila-share-1-1G': {'size': 1},
             'root@host2:/manila-share-2-2G': {'size': 2}})
        self.glusterfs_used_vols = set([
            'root@host1:/manila-share-1-1G',
            'root@host2:/manila-share-2-2G'])

        CONF.set_default('glusterfs_volume_pattern',
                         'manila-share-\d+-#{size}G$')
        CONF.set_default('driver_handles_share_servers', False)

        self.fake_conf = config.Configuration(None)
        self.mock_object(common.GlusterManager, 'make_gluster_call')

        self._driver = glusterfs_native.GlusterfsNativeShareDriver(
            execute=self._execute,
            configuration=self.fake_conf)
        self.addCleanup(fake_utils.fake_execute_set_repliers, [])
        self.addCleanup(fake_utils.fake_execute_clear_log)

    def test_supported_protocols(self):
        self.assertEqual(('GLUSTERFS', ),
                         self._driver.supported_protocols)

    def test_setup_via_manager(self):
        gmgr = mock.Mock()

        gmgr.gluster_call = mock.Mock()
        gmgr.set_vol_option = mock.Mock()
        gmgr.volume = 'fakevol'
        gmgr.export = 'fakehost:/fakevol'
        gmgr.get_vol_option = mock.Mock(
            return_value='glusterfs-server-name,some-other-name')
        share = mock.Mock()
        settings = (
            ('nfs.export-volumes', False, {}),
            ('client.ssl', True, {}),
            ('server.ssl', True, {}),
            ('server.dynamic-auth', True, {'ignore_failure': True}),
        )
        call_args = (
            ('volume', 'stop', 'fakevol', '--mode=script', {'log': mock.ANY}),
            ('volume', 'start', 'fakevol', {'log': mock.ANY}),
        )

        ret = self._driver._setup_via_manager({'manager': gmgr,
                                               'share': share})

        gmgr.get_vol_option.assert_called_once_with('auth.ssl-allow')
        gmgr.set_vol_option.assert_has_calls(
            [mock.call(*a[:-1], **a[-1]) for a in settings])
        gmgr.gluster_call.assert_has_calls(
            [mock.call(*a[:-1], **a[-1]) for a in call_args])
        self.assertEqual(ret, gmgr.export)

    def test_setup_via_manager_with_parent(self):
        gmgr = mock.Mock()
        gmgr.set_vol_option = mock.Mock()
        gmgr.volume = 'fakevol'
        gmgr.export = 'fakehost:/fakevol'
        gmgr_parent = mock.Mock()
        gmgr_parent.get_vol_option = mock.Mock(
            return_value=(
                'glusterfs-server-name,some-other-name,manila-host.com'))
        share = mock.Mock()
        share_parent = mock.Mock()
        settings = (
            ('auth.ssl-allow',
             'glusterfs-server-name,manila-host.com', {}),
            ('nfs.export-volumes', False, {}),
            ('client.ssl', True, {}),
            ('server.ssl', True, {}),
            ('server.dynamic-auth', True, {'ignore_failure': True}),
        )

        ret = self._driver._setup_via_manager(
            {'manager': gmgr, 'share': share},
            {'manager': gmgr_parent, 'share': share_parent})

        gmgr_parent.get_vol_option.assert_called_once_with(
            'auth.ssl-allow')
        gmgr.set_vol_option.assert_has_calls(
            [mock.call(*a[:-1], **a[-1]) for a in settings])
        self.assertEqual(ret, gmgr.export)

    @ddt.data(True, False)
    def test_setup_via_manager_no_option_data(self, has_parent):
        share = mock.Mock()
        gmgr = mock.Mock()
        if has_parent:
            share_parent = mock.Mock()
            gmgr_parent = mock.Mock()
            share_mgr_parent = {'share': share_parent, 'manager': gmgr_parent}
            gmgr_queried = gmgr_parent
        else:
            share_mgr_parent = None
            gmgr_queried = gmgr
        gmgr_queried.get_vol_option = mock.Mock(return_value='')

        self.assertRaises(exception.GlusterfsException,
                          self._driver._setup_via_manager,
                          {'share': share, 'manager': gmgr},
                          share_mgr_parent=share_mgr_parent)

        gmgr_queried.get_vol_option.assert_called_once_with(
            'auth.ssl-allow')

    def test_snapshots_are_supported(self):
        self.assertTrue(self._driver.snapshots_are_supported)

    @ddt.data({'delta': (["oldCN"], []), 'expected': "glusterCN,oldCN"},
              {'delta': (["newCN"], []), 'expected': "glusterCN,newCN,oldCN"},
              {'delta': ([], ["newCN"]), 'expected': "glusterCN,oldCN"},
              {'delta': ([], ["oldCN"]), 'expected': "glusterCN"})
    @ddt.unpack
    def test_update_access_via_manager(self, delta, expected):
        gluster_mgr = common.GlusterManager(self.glusterfs_target1,
                                            self._execute, None, None)
        self.mock_object(
            gluster_mgr, 'get_vol_option', mock.Mock(
                side_effect=lambda a, *x, **kw: {
                    'auth.ssl-allow': "glusterCN,oldCN",
                    'server.dynamic-auth': True}[a]))
        self.mock_object(gluster_mgr, 'set_vol_option')
        add_rules, delete_rules = (
            map(lambda a: {'access_to': a}, r) for r in delta)

        self._driver._update_access_via_manager(
            gluster_mgr, self._context, self.share1, add_rules, delete_rules)

        argseq = [('auth.ssl-allow', {})]
        if delete_rules:
            argseq.append(('server.dynamic-auth', {'boolean': True}))
        self.assertEqual([mock.call(a[0], **a[1]) for a in argseq],
                         gluster_mgr.get_vol_option.call_args_list)
        gluster_mgr.set_vol_option.assert_called_once_with('auth.ssl-allow',
                                                           expected)

    def test_update_access_via_manager_restart(self):
        gluster_mgr = common.GlusterManager(self.glusterfs_target1,
                                            self._execute, None, None)
        self.mock_object(
            gluster_mgr, 'get_vol_option', mock.Mock(
                side_effect=lambda a, *x, **kw: {
                    'auth.ssl-allow': "glusterCN,oldCN",
                    'server.dynamic-auth': False}[a]))
        self.mock_object(gluster_mgr, 'set_vol_option')
        self.mock_object(common, '_restart_gluster_vol')

        self._driver._update_access_via_manager(
            gluster_mgr, self._context, self.share1, [],
            [{'access_to': "oldCN"}])

        common._restart_gluster_vol.assert_called_once_with(gluster_mgr)

    @ddt.data('common name with space', 'comma,nama')
    def test_update_access_via_manager_badcn(self, common_name):
        gluster_mgr = common.GlusterManager(self.glusterfs_target1,
                                            self._execute, None, None)
        self.mock_object(gluster_mgr, 'get_vol_option', mock.Mock(
            return_value="glusterCN,oldCN"))

        self.assertRaises(exception.GlusterfsException,
                          self._driver._update_access_via_manager,
                          gluster_mgr, self._context, self.share1,
                          [{'access_to': common_name}], [])

    def test_update_share_stats(self):
        self._driver._update_share_stats()

        test_data = {
            'share_backend_name': 'GlusterFS-Native',
            'driver_handles_share_servers': False,
            'vendor_name': 'Red Hat',
            'driver_version': '1.1',
            'storage_protocol': 'glusterfs',
            'reserved_percentage': 0,
            'qos': False,
            'total_capacity_gb': 'unknown',
            'free_capacity_gb': 'unknown',
            'pools': None,
            'snapshot_support': True,
            'create_share_from_snapshot_support': True,
            'revert_to_snapshot_support': False,
            'mount_snapshot_support': False,
            'share_group_stats': {
                'consistent_snapshot_support': None,
            },
            'replication_domain': None,
            'filter_function': None,
            'goodness_function': None,
            'ipv4_support': True,
            'ipv6_support': False,
        }
        self.assertEqual(test_data, self._driver._stats)

    def test_get_network_allocations_number(self):
        self.assertEqual(0, self._driver.get_network_allocations_number())



# Copyright 2012, Intel, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Client side of the share RPC API.
"""

from oslo_config import cfg
import oslo_messaging as messaging
from oslo_serialization import jsonutils

from manila import rpc
from manila.share import utils

CONF = cfg.CONF


class ShareAPI(object):
    """Client side of the share rpc API.

    API version history:

        1.0  - Initial version.
        1.1  - Add manage_share() and unmanage_share() methods
        1.2  - Add extend_share() method
        1.3  - Add shrink_share() method
        1.4  - Introduce Share Instances:
            create_share() -> create_share_instance()
            delete_share() -> delete_share_instance()
            Add share_instance argument to allow_access() & deny_access()
        1.5  - Add create_consistency_group, delete_consistency_group
                create_cgsnapshot, and delete_cgsnapshot methods
        1.6  - Introduce Share migration:
            migrate_share()
            get_migration_info()
            get_driver_migration_info()
        1.7  - Update target call API in allow/deny access methods (Removed
            in 1.14)
        1.8  - Introduce Share Replication:
            create_share_replica()
            delete_share_replica()
            promote_share_replica()
            update_share_replica()
        1.9  - Add manage_snapshot() and unmanage_snapshot() methods
        1.10 - Add migration_complete(), migration_cancel() and
            migration_get_progress(), rename migrate_share() to
            migration_start(), rename get_migration_info() to
            migration_get_info(), rename get_driver_migration_info() to
            migration_get_driver_info()
        1.11 - Add create_replicated_snapshot() and
            delete_replicated_snapshot() methods
        1.12 - Add provide_share_server(), create_share_server() and
            migration_driver_recovery(), remove migration_get_driver_info(),
            update migration_cancel(), migration_complete() and
            migration_get_progress method signature, rename
            migration_get_info() to connection_get_info()
        1.13 - Introduce share revert to snapshot: revert_to_snapshot()
        1.14 - Add update_access() and remove allow_access() and deny_access().
        1.15 - Updated migration_start() method with new parameter
            "preserve_snapshots"
        1.16  - Convert create_consistency_group, delete_consistency_group
                create_cgsnapshot, and delete_cgsnapshot methods to
                create_share_group, delete_share_group
                create_share_group_snapshot, and delete_share_group_snapshot
        1.17 - Add snapshot_update_access()
        1.18 - Remove unused "share_id" parameter from revert_to_snapshot()
        1.19 - Add manage_share_server() and unmanage_share_server()
    """

    BASE_RPC_API_VERSION = '1.0'

    def __init__(self, topic=None):
        super(ShareAPI, self).__init__()
        target = messaging.Target(topic=CONF.share_topic,
                                  version=self.BASE_RPC_API_VERSION)
        self.client = rpc.get_client(target, version_cap='1.19')

    def create_share_instance(self, context, share_instance, host,
                              request_spec, filter_properties,
                              snapshot_id=None):
        new_host = utils.extract_host(host)
        call_context = self.client.prepare(server=new_host, version='1.4')
        request_spec_p = jsonutils.to_primitive(request_spec)
        call_context.cast(context,
                          'create_share_instance',
                          share_instance_id=share_instance['id'],
                          request_spec=request_spec_p,
                          filter_properties=filter_properties,
                          snapshot_id=snapshot_id)

    def manage_share(self, context, share, driver_options=None):
        host = utils.extract_host(share['instance']['host'])
        call_context = self.client.prepare(server=host, version='1.1')
        call_context.cast(context,
                          'manage_share',
                          share_id=share['id'],
                          driver_options=driver_options)

    def unmanage_share(self, context, share):
        host = utils.extract_host(share['instance']['host'])
        call_context = self.client.prepare(server=host, version='1.1')
        call_context.cast(context, 'unmanage_share', share_id=share['id'])

    def manage_snapshot(self, context, snapshot, host,
                        driver_options=None):
        new_host = utils.extract_host(host)
        call_context = self.client.prepare(server=new_host, version='1.9')
        call_context.cast(context,
                          'manage_snapshot',
                          snapshot_id=snapshot['id'],
                          driver_options=driver_options)

    def unmanage_snapshot(self, context, snapshot, host):
        new_host = utils.extract_host(host)
        call_context = self.client.prepare(server=new_host, version='1.9')
        call_context.cast(context,
                          'unmanage_snapshot',
                          snapshot_id=snapshot['id'])

    def manage_share_server(
            self, context, share_server, identifier, driver_opts):
        host = utils.extract_host(share_server['host'])
        call_context = self.client.prepare(server=host, version='1.19')
        call_context.cast(context, 'manage_share_server',
                          share_server_id=share_server['id'],
                          identifier=identifier,
                          driver_opts=driver_opts)

    def unmanage_share_server(self, context, share_server, force=False):
        host = utils.extract_host(share_server['host'])
        call_context = self.client.prepare(server=host, version='1.19')
        call_context.cast(context, 'unmanage_share_server',
                          share_server_id=share_server['id'],
                          force=force)

    def revert_to_snapshot(self, context, share, snapshot, host, reservations):
        host = utils.extract_host(host)
        call_context = self.client.prepare(server=host, version='1.18')
        call_context.cast(context,
                          'revert_to_snapshot',
                          snapshot_id=snapshot['id'],
                          reservations=reservations)

    def delete_share_instance(self, context, share_instance, force=False):
        host = utils.extract_host(share_instance['host'])
        call_context = self.client.prepare(server=host, version='1.4')
        call_context.cast(context,
                          'delete_share_instance',
                          share_instance_id=share_instance['id'],
                          force=force)

    def migration_start(self, context, share, dest_host,
                        force_host_assisted_migration, preserve_metadata,
                        writable, nondisruptive, preserve_snapshots,
                        new_share_network_id, new_share_type_id):
        new_host = utils.extract_host(share['instance']['host'])
        call_context = self.client.prepare(server=new_host, version='1.15')
        call_context.cast(
            context,
            'migration_start',
            share_id=share['id'],
            dest_host=dest_host,
            force_host_assisted_migration=force_host_assisted_migration,
            preserve_metadata=preserve_metadata,
            writable=writable,
            nondisruptive=nondisruptive,
            preserve_snapshots=preserve_snapshots,
            new_share_network_id=new_share_network_id,
            new_share_type_id=new_share_type_id)

    def connection_get_info(self, context, share_instance):
        new_host = utils.extract_host(share_instance['host'])
        call_context = self.client.prepare(server=new_host, version='1.12')
        return call_context.call(context,
                                 'connection_get_info',
                                 share_instance_id=share_instance['id'])

    def delete_share_server(self, context, share_server):
        host = utils.extract_host(share_server['host'])
        call_context = self.client.prepare(server=host, version='1.0')
        call_context.cast(context,
                          'delete_share_server',
                          share_server=share_server)

    def create_snapshot(self, context, share, snapshot):
        host = utils.extract_host(share['instance']['host'])
        call_context = self.client.prepare(server=host)
        call_context.cast(context,
                          'create_snapshot',
                          share_id=share['id'],
                          snapshot_id=snapshot['id'])

    def delete_snapshot(self, context, snapshot, host, force=False):
        new_host = utils.extract_host(host)
        call_context = self.client.prepare(server=new_host)
        call_context.cast(context,
                          'delete_snapshot',
                          snapshot_id=snapshot['id'],
                          force=force)

    def create_replicated_snapshot(self, context, share, replicated_snapshot):
        host = utils.extract_host(share['instance']['host'])
        call_context = self.client.prepare(server=host, version='1.11')
        call_context.cast(context,
                          'create_replicated_snapshot',
                          snapshot_id=replicated_snapshot['id'],
                          share_id=share['id'])

    def delete_replicated_snapshot(self, context, replicated_snapshot, host,
                                   share_id=None, force=False):
        host = utils.extract_host(host)
        call_context = self.client.prepare(server=host, version='1.11')
        call_context.cast(context,
                          'delete_replicated_snapshot',
                          snapshot_id=replicated_snapshot['id'],
                          share_id=share_id,
                          force=force)

    def update_access(self, context, share_instance):
        host = utils.extract_host(share_instance['host'])
        call_context = self.client.prepare(server=host, version='1.14')
        call_context.cast(context, 'update_access',
                          share_instance_id=share_instance['id'])

    def publish_service_capabilities(self, context):
        call_context = self.client.prepare(fanout=True, version='1.0')
        call_context.cast(context, 'publish_service_capabilities')

    def extend_share(self, context, share, new_size, reservations):
        host = utils.extract_host(share['instance']['host'])
        call_context = self.client.prepare(server=host, version='1.2')
        call_context.cast(context,
                          'extend_share',
                          share_id=share['id'],
                          new_size=new_size,
                          reservations=reservations)

    def shrink_share(self, context, share, new_size):
        host = utils.extract_host(share['instance']['host'])
        call_context = self.client.prepare(server=host, version='1.3')
        call_context.cast(context,
                          'shrink_share',
                          share_id=share['id'],
                          new_size=new_size)

    def create_share_group(self, context, share_group, host):
        new_host = utils.extract_host(host)
        call_context = self.client.prepare(server=new_host, version='1.16')
        call_context.cast(
            context, 'create_share_group', share_group_id=share_group['id'])

    def delete_share_group(self, context, share_group):
        new_host = utils.extract_host(share_group['host'])
        call_context = self.client.prepare(server=new_host, version='1.16')
        call_context.cast(
            context, 'delete_share_group', share_group_id=share_group['id'])

    def create_share_group_snapshot(self, context, share_group_snapshot, host):
        new_host = utils.extract_host(host)
        call_context = self.client.prepare(server=new_host, version='1.16')
        call_context.cast(
            context, 'create_share_group_snapshot',
            share_group_snapshot_id=share_group_snapshot['id'])

    def delete_share_group_snapshot(self, context, share_group_snapshot, host):
        new_host = utils.extract_host(host)
        call_context = self.client.prepare(server=new_host, version='1.16')
        call_context.cast(
            context, 'delete_share_group_snapshot',
            share_group_snapshot_id=share_group_snapshot['id'])

    def create_share_replica(self, context, share_replica, host,
                             request_spec, filter_properties):
        new_host = utils.extract_host(host)
        call_context = self.client.prepare(server=new_host, version='1.8')
        request_spec_p = jsonutils.to_primitive(request_spec)
        call_context.cast(context,
                          'create_share_replica',
                          share_replica_id=share_replica['id'],
                          request_spec=request_spec_p,
                          filter_properties=filter_properties,
                          share_id=share_replica['share_id'])

    def delete_share_replica(self, context, share_replica, force=False):
        host = utils.extract_host(share_replica['host'])
        call_context = self.client.prepare(server=host, version='1.8')
        call_context.cast(context,
                          'delete_share_replica',
                          share_replica_id=share_replica['id'],
                          share_id=share_replica['share_id'],
                          force=force)

    def promote_share_replica(self, context, share_replica):
        host = utils.extract_host(share_replica['host'])
        call_context = self.client.prepare(server=host, version='1.8')
        call_context.cast(context,
                          'promote_share_replica',
                          share_replica_id=share_replica['id'],
                          share_id=share_replica['share_id'])

    def update_share_replica(self, context, share_replica):
        host = utils.extract_host(share_replica['host'])
        call_context = self.client.prepare(server=host, version='1.8')
        call_context.cast(context,
                          'update_share_replica',
                          share_replica_id=share_replica['id'],
                          share_id=share_replica['share_id'])

    def migration_complete(self, context, src_share_instance,
                           dest_instance_id):
        new_host = utils.extract_host(src_share_instance['host'])
        call_context = self.client.prepare(server=new_host, version='1.12')
        call_context.cast(context,
                          'migration_complete',
                          src_instance_id=src_share_instance['id'],
                          dest_instance_id=dest_instance_id)

    def migration_cancel(self, context, src_share_instance, dest_instance_id):
        new_host = utils.extract_host(src_share_instance['host'])
        call_context = self.client.prepare(server=new_host, version='1.12')
        call_context.cast(context,
                          'migration_cancel',
                          src_instance_id=src_share_instance['id'],
                          dest_instance_id=dest_instance_id)

    def migration_get_progress(self, context, src_share_instance,
                               dest_instance_id):
        new_host = utils.extract_host(src_share_instance['host'])
        call_context = self.client.prepare(server=new_host, version='1.12')
        return call_context.call(context,
                                 'migration_get_progress',
                                 src_instance_id=src_share_instance['id'],
                                 dest_instance_id=dest_instance_id)

    def provide_share_server(self, context, share_instance, share_network_id,
                             snapshot_id=None):
        new_host = utils.extract_host(share_instance['host'])
        call_context = self.client.prepare(server=new_host, version='1.12')
        return call_context.call(context,
                                 'provide_share_server',
                                 share_instance_id=share_instance['id'],
                                 share_network_id=share_network_id,
                                 snapshot_id=snapshot_id)

    def create_share_server(self, context, share_instance, share_server_id):
        new_host = utils.extract_host(share_instance['host'])
        call_context = self.client.prepare(server=new_host, version='1.12')
        call_context.cast(context,
                          'create_share_server',
                          share_server_id=share_server_id)

    def snapshot_update_access(self, context, snapshot_instance):
        host = utils.extract_host(snapshot_instance['share_instance']['host'])
        call_context = self.client.prepare(server=host, version='1.17')
        call_context.cast(context,
                          'snapshot_update_access',
                          snapshot_instance_id=snapshot_instance['id'])



# Copyright 2017 Infinidat Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
INFINIDAT InfiniBox Share Driver
"""

import functools

import ipaddress
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import units
import six

from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.share import driver
from manila.share import utils
from manila import version

try:
    import capacity
    import infinisdk
except ImportError:
    capacity = None
    infinisdk = None


LOG = logging.getLogger(__name__)

infinidat_connection_opts = [
    cfg.HostAddressOpt('infinibox_hostname',
                       help='The name (or IP address) for the INFINIDAT '
                       'Infinibox storage system.'), ]

infinidat_auth_opts = [
    cfg.StrOpt('infinibox_login',
               help=('Administrative user account name used to access the '
                     'INFINIDAT Infinibox storage system.')),
    cfg.StrOpt('infinibox_password',
               help=('Password for the administrative user account '
                     'specified in the infinibox_login option.'),
               secret=True), ]

infinidat_general_opts = [
    cfg.StrOpt('infinidat_pool_name',
               help='Name of the pool from which volumes are allocated.'),
    cfg.StrOpt('infinidat_nas_network_space_name',
               help='Name of the NAS network space on the INFINIDAT '
               'InfiniBox.'),
    cfg.BoolOpt('infinidat_thin_provision', help='Use thin provisioning.',
                default=True)]

CONF = cfg.CONF
CONF.register_opts(infinidat_connection_opts)
CONF.register_opts(infinidat_auth_opts)
CONF.register_opts(infinidat_general_opts)

_MANILA_TO_INFINIDAT_ACCESS_LEVEL = {
    constants.ACCESS_LEVEL_RW: 'RW',
    constants.ACCESS_LEVEL_RO: 'RO',
}

# Max retries for the REST API client in case of a failure:
_API_MAX_RETRIES = 5
# Identifier used as the REST API User-Agent string:
_INFINIDAT_MANILA_IDENTIFIER = (
    "manila/%s" % version.version_info.release_string())


def infinisdk_to_manila_exceptions(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except infinisdk.core.exceptions.InfiniSDKException as ex:
            # string formatting of 'ex' includes http code and url
            msg = _('Caught exception from infinisdk: %s') % ex
            LOG.exception(msg)
            raise exception.ShareBackendException(msg=msg)
    return wrapper


class InfiniboxShareDriver(driver.ShareDriver):

    VERSION = '1.0'    # driver version

    def __init__(self, *args, **kwargs):
        super(InfiniboxShareDriver, self).__init__(False, *args, **kwargs)
        self.configuration.append_config_values(infinidat_connection_opts)
        self.configuration.append_config_values(infinidat_auth_opts)
        self.configuration.append_config_values(infinidat_general_opts)

    def _setup_and_get_system_object(self, management_address, auth):
        system = infinisdk.InfiniBox(management_address, auth=auth)
        system.api.add_auto_retry(
            lambda e: isinstance(
                e, infinisdk.core.exceptions.APITransportFailure) and
            "Interrupted system call" in e.error_desc, _API_MAX_RETRIES)
        system.api.set_source_identifier(_INFINIDAT_MANILA_IDENTIFIER)
        system.login()
        return system

    def do_setup(self, context):
        """Driver initialization"""
        if infinisdk is None:
            msg = _("Missing 'infinisdk' python module, ensure the library"
                    " is installed and available.")
            raise exception.ManilaException(message=msg)

        infinibox_login = self._safe_get_from_config_or_fail('infinibox_login')
        infinibox_password = (
            self._safe_get_from_config_or_fail('infinibox_password'))
        auth = (infinibox_login, infinibox_password)

        management_address = (
            self._safe_get_from_config_or_fail('infinibox_hostname'))

        self._pool_name = (
            self._safe_get_from_config_or_fail('infinidat_pool_name'))

        self._network_space_name = (
            self._safe_get_from_config_or_fail(
                'infinidat_nas_network_space_name'))

        self._system = (
            self._setup_and_get_system_object(management_address, auth))

        backend_name = self.configuration.safe_get('share_backend_name')
        self._backend_name = backend_name or self.__class__.__name__

        thin_provisioning = self.configuration.infinidat_thin_provision
        self._provtype = "THIN" if thin_provisioning else "THICK"

        LOG.debug('setup complete')

    def _update_share_stats(self):
        """Retrieve stats info from share group."""
        (free_capacity_bytes, physical_capacity_bytes,
         provisioned_capacity_gb) = self._get_available_capacity()

        max_over_subscription_ratio = (
            self.configuration.max_over_subscription_ratio)

        data = dict(
            share_backend_name=self._backend_name,
            vendor_name='INFINIDAT',
            driver_version=self.VERSION,
            storage_protocol='NFS',
            total_capacity_gb=float(physical_capacity_bytes) / units.Gi,
            free_capacity_gb=float(free_capacity_bytes) / units.Gi,
            reserved_percentage=self.configuration.reserved_share_percentage,
            thin_provisioning=self.configuration.infinidat_thin_provision,
            max_over_subscription_ratio=max_over_subscription_ratio,
            provisioned_capacity_gb=provisioned_capacity_gb,
            snapshot_support=True,
            create_share_from_snapshot_support=True,
            mount_snapshot_support=True,
            revert_to_snapshot_support=True)

        super(InfiniboxShareDriver, self)._update_share_stats(data)

    def _get_available_capacity(self):
        # pylint: disable=no-member
        pool = self._get_infinidat_pool()
        free_capacity_bytes = (pool.get_free_physical_capacity() /
                               capacity.byte)
        physical_capacity_bytes = (pool.get_physical_capacity() /
                                   capacity.byte)
        provisioned_capacity_gb = (
            (pool.get_virtual_capacity() - pool.get_free_virtual_capacity()) /
            capacity.GB)
        # pylint: enable=no-member
        return (free_capacity_bytes, physical_capacity_bytes,
                provisioned_capacity_gb)

    def _safe_get_from_config_or_fail(self, config_parameter):
        config_value = self.configuration.safe_get(config_parameter)
        if not config_value:    # None or empty string
            reason = (_("%(config_parameter)s configuration parameter "
                        "must be specified") %
                      {'config_parameter': config_parameter})
            LOG.error(reason)
            raise exception.BadConfigurationException(reason=reason)
        return config_value

    def _verify_share_protocol(self, share):
        if share['share_proto'] != 'NFS':
            reason = (_('Unsupported share protocol: %(proto)s.') %
                      {'proto': share['share_proto']})
            LOG.error(reason)
            raise exception.InvalidShare(reason=reason)

    def _verify_access_type(self, access):
        if access['access_type'] != 'ip':
            reason = _('Only "ip" access type allowed for the NFS protocol.')
            LOG.error(reason)
            raise exception.InvalidShareAccess(reason=reason)
        return True

    def _make_share_name(self, manila_share):
        return 'openstack-shr-%s' % manila_share['id']

    def _make_snapshot_name(self, manila_snapshot):
        return 'openstack-snap-%s' % manila_snapshot['id']

    def _set_manila_object_metadata(self, infinidat_object, manila_object):
        data = {"system": "openstack",
                "openstack_version": version.version_info.release_string(),
                "manila_id": manila_object['id'],
                "manila_name": manila_object['name'],
                "host.created_by": _INFINIDAT_MANILA_IDENTIFIER}
        infinidat_object.set_metadata_from_dict(data)

    @infinisdk_to_manila_exceptions
    def _get_infinidat_pool(self):
        pool = self._system.pools.safe_get(name=self._pool_name)
        if pool is None:
            msg = _('Pool "%s" not found') % self._pool_name
            LOG.error(msg)
            raise exception.ShareBackendException(msg=msg)
        return pool

    @infinisdk_to_manila_exceptions
    def _get_infinidat_nas_network_space_ips(self):
        network_space = self._system.network_spaces.safe_get(
            name=self._network_space_name)
        if network_space is None:
            msg = _('INFINIDAT InfiniBox NAS network space "%s" '
                    'not found') % self._network_space_name
            LOG.error(msg)
            raise exception.ShareBackendException(msg=msg)
        network_space_ips = network_space.get_ips()
        if not network_space_ips:
            msg = _('INFINIDAT InfiniBox NAS network space "%s" has no IP '
                    'addresses defined') % self._network_space_name
            LOG.error(msg)
            raise exception.ShareBackendException(msg=msg)
        ip_addresses = (
            [ip_munch.ip_address for ip_munch in network_space_ips if
             ip_munch.enabled])
        if not ip_addresses:
            msg = _('INFINIDAT InfiniBox NAS network space "%s" has no '
                    'enabled IP addresses') % self._network_space_name
            LOG.error(msg)
            raise exception.ShareBackendException(msg=msg)
        return ip_addresses

    def _get_full_nfs_export_paths(self, export_path):
        network_space_ips = self._get_infinidat_nas_network_space_ips()
        return ['{network_space_ip}:{export_path}'.format(
            network_space_ip=network_space_ip,
            export_path=export_path) for network_space_ip in network_space_ips]

    @infinisdk_to_manila_exceptions
    def _get_infinidat_filesystem_by_name(self, name):
        filesystem = self._system.filesystems.safe_get(name=name)
        if filesystem is None:
            msg = (_('Filesystem not found on the Infinibox by its name: %s') %
                   name)
            LOG.error(msg)
            raise exception.ShareResourceNotFound(share_id=name)
        return filesystem

    def _get_infinidat_filesystem(self, manila_share):
        filesystem_name = self._make_share_name(manila_share)
        return self._get_infinidat_filesystem_by_name(filesystem_name)

    def _get_infinidat_snapshot_by_name(self, name):
        snapshot = self._system.filesystems.safe_get(name=name)
        if snapshot is None:
            msg = (_('Snapshot not found on the Infinibox by its name: %s') %
                   name)
            LOG.error(msg)
            raise exception.ShareSnapshotNotFound(snapshot_id=name)
        return snapshot

    def _get_infinidat_snapshot(self, manila_snapshot):
        snapshot_name = self._make_snapshot_name(manila_snapshot)
        return self._get_infinidat_snapshot_by_name(snapshot_name)

    def _get_infinidat_dataset(self, manila_object, is_snapshot):
        return (self._get_infinidat_snapshot(manila_object) if is_snapshot
                else self._get_infinidat_filesystem(manila_object))

    @infinisdk_to_manila_exceptions
    def _get_export(self, infinidat_filesystem):
        infinidat_exports = infinidat_filesystem.get_exports()
        if len(infinidat_exports) == 0:
            msg = _("Could not find share export")
            raise exception.ShareBackendException(msg=msg)
        elif len(infinidat_exports) > 1:
            msg = _("INFINIDAT filesystem has more than one active export; "
                    "possibly not a Manila share")
            LOG.error(msg)
            raise exception.ShareBackendException(msg=msg)
        return infinidat_exports[0]

    def _get_infinidat_access_level(self, access):
        """Translates between Manila access levels to INFINIDAT API ones"""
        access_level = access['access_level']
        try:
            return _MANILA_TO_INFINIDAT_ACCESS_LEVEL[access_level]
        except KeyError:
            raise exception.InvalidShareAccessLevel(level=access_level)

    def _get_ip_address_range(self, ip_address):
        """Parse single IP address or subnet into a range.

        If the IP address string is in subnet mask format, returns a
        <start ip>-<end-ip> string. If the IP address contains a single IP
        address, returns only that IP address.
        """

        ip_address = six.text_type(ip_address)

        # try treating the ip_address parameter as a range of IP addresses:
        ip_network = ipaddress.ip_network(ip_address, strict=False)
        ip_network_hosts = list(ip_network.hosts())
        if len(ip_network_hosts) == 0:    # /32, single IP address
            return ip_address.split('/')[0]
        return "{}-{}".format(ip_network_hosts[0], ip_network_hosts[-1])

    @infinisdk_to_manila_exceptions
    def _create_filesystem_export(self, infinidat_filesystem):
        infinidat_export = infinidat_filesystem.add_export(permissions=[])
        export_paths = self._get_full_nfs_export_paths(
            infinidat_export.get_export_path())
        export_locations = [{
            'path': export_path,
            'is_admin_only': False,
            'metadata': {},
        } for export_path in export_paths]
        return export_locations

    @infinisdk_to_manila_exceptions
    def _delete_share(self, share, is_snapshot):
        if is_snapshot:
            dataset_name = self._make_snapshot_name(share)
        else:
            dataset_name = self._make_share_name(share)
        try:
            infinidat_filesystem = (
                self._get_infinidat_filesystem_by_name(dataset_name))
        except exception.ShareResourceNotFound:
            message = ("share %(share)s not found on Infinibox, skipping "
                       "delete")
            LOG.warning(message, {"share": share})
            return      # filesystem not found
        try:
            infinidat_export = self._get_export(infinidat_filesystem)
            infinidat_export.safe_delete()
        except exception.ShareBackendException:
            # it is possible that the export has been deleted
            pass
        infinidat_filesystem.safe_delete()

    @infinisdk_to_manila_exceptions
    def _extend_share(self, infinidat_filesystem, share, new_size):
        # pylint: disable=no-member
        new_size_capacity_units = new_size * capacity.GiB
        # pylint: enable=no-member
        old_size = infinidat_filesystem.get_size()
        infinidat_filesystem.resize(new_size_capacity_units - old_size)

    @infinisdk_to_manila_exceptions
    def _update_access(self, manila_object, access_rules, is_snapshot):
        infinidat_filesystem = self._get_infinidat_dataset(
            manila_object, is_snapshot=is_snapshot)
        infinidat_export = self._get_export(infinidat_filesystem)
        permissions = [
            {'access': self._get_infinidat_access_level(access_rule),
             'client': self._get_ip_address_range(access_rule['access_to']),
             'no_root_squash': True} for access_rule in access_rules if
            self._verify_access_type(access_rule)]
        infinidat_export.update_permissions(permissions)

    @infinisdk_to_manila_exceptions
    def create_share(self, context, share, share_server=None):
        self._verify_share_protocol(share)

        pool = self._get_infinidat_pool()
        size = share['size'] * capacity.GiB    # pylint: disable=no-member
        share_name = self._make_share_name(share)

        infinidat_filesystem = self._system.filesystems.create(
            pool=pool, name=share_name, size=size, provtype=self._provtype)
        self._set_manila_object_metadata(infinidat_filesystem, share)
        return self._create_filesystem_export(infinidat_filesystem)

    @infinisdk_to_manila_exceptions
    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None):
        name = self._make_share_name(share)
        infinidat_snapshot = self._get_infinidat_snapshot(snapshot)
        infinidat_new_share = infinidat_snapshot.create_snapshot(
            name=name, write_protected=False)
        self._extend_share(infinidat_new_share, share, share['size'])
        return self._create_filesystem_export(infinidat_new_share)

    @infinisdk_to_manila_exceptions
    def create_snapshot(self, context, snapshot, share_server=None):
        """Creates a snapshot."""
        share = snapshot['share']
        infinidat_filesystem = self._get_infinidat_filesystem(share)
        name = self._make_snapshot_name(snapshot)
        infinidat_snapshot = infinidat_filesystem.create_snapshot(name=name)
        # snapshot is created in the same size as the original share, so no
        # extending is needed
        self._set_manila_object_metadata(infinidat_snapshot, snapshot)
        return {'export_locations':
                self._create_filesystem_export(infinidat_snapshot)}

    def delete_share(self, context, share, share_server=None):
        try:
            self._verify_share_protocol(share)
        except exception.InvalidShare:
            # cleanup shouldn't fail on wrong protocol or missing share:
            message = ("failed to delete share %(share)s; unsupported share "
                       "protocol %(share_proto)s, only NFS is supported")
            LOG.warning(message, {"share": share,
                        "share_proto": share['share_proto']})
            return
        self._delete_share(share, is_snapshot=False)

    def delete_snapshot(self, context, snapshot, share_server=None):
        self._delete_share(snapshot, is_snapshot=True)

    def ensure_share(self, context, share, share_server=None):
        # will raise ShareResourceNotFound if the share was not found:
        infinidat_filesystem = self._get_infinidat_filesystem(share)
        try:
            infinidat_export = self._get_export(infinidat_filesystem)
            return self._get_full_nfs_export_paths(
                infinidat_export.get_export_path())
        except exception.ShareBackendException:
            # export not found, need to re-export
            message = ("missing export for share %(share)s, trying to "
                       "re-export")
            LOG.info(message, {"share": share})
            return self._create_filesystem_export(infinidat_filesystem)

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        # As the Infinibox API can bulk update export access rules, we will try
        # to use the access_rules list
        self._verify_share_protocol(share)
        self._update_access(share, access_rules, is_snapshot=False)

    def get_network_allocations_number(self):
        return 0

    @infinisdk_to_manila_exceptions
    def revert_to_snapshot(self, context, snapshot, share_access_rules,
                           snapshot_access_rules, share_server=None):
        infinidat_snapshot = self._get_infinidat_snapshot(snapshot)
        infinidat_parent_share = self._get_infinidat_filesystem(
            snapshot['share'])
        infinidat_parent_share.restore(infinidat_snapshot)

    def extend_share(self, share, new_size, share_server=None):
        infinidat_filesystem = self._get_infinidat_filesystem(share)
        self._extend_share(infinidat_filesystem, share, new_size)

    def snapshot_update_access(self, context, snapshot, access_rules,
                               add_rules, delete_rules, share_server=None):
        # snapshots are to be mounted in read-only mode, see:
        # "Add mountable snapshots" on openstack specs.
        access_rules, _, _ = utils.change_rules_to_readonly(
            access_rules, [], [])
        try:
            self._update_access(snapshot, access_rules, is_snapshot=True)
        except exception.InvalidShareAccess as e:
            raise exception.InvalidSnapshotAccess(e)
